# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


# Standard Library Imports
import logging
from tkinter import messagebox

# Third Party Imports

# Local Imports
from navigate.controller.sub_controllers.gui_controller import GUIController
from navigate.view.popups.acquire_popup import AcquirePopUp

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class AcquireBarController(GUIController):
    """Acquire Bar Controller."""

    def __init__(self, view, parent_controller):
        """Initialize the Acquire Bar Controller.

        Parameters
        ----------
        view : object
            Instance of the Acquire Bar View.
        parent_view : object
            Instance of the Main View.
        parent_controller : object
            Instance of the Main Controller.
        """
        super().__init__(view, parent_controller)

        #: str: Acquisition image mode.
        self.mode = "live"
        #: bool: Whether the image will be saved.
        self.is_save = False
        #: bool: Whether the microscope is acquiring.
        self.is_acquiring = False
        #: dict: Dictionary of different operating modes.
        self.mode_dict = {
            "Continuous Scan": "live",
            "Z-Stack": "z-stack",
            "Single Acquisition": "single",
            "Customized": "customized",
            "Constant Velocity Acquisition": "ConstantVelocityAcquisition",
        }

        self.view.pull_down["values"] = list(self.mode_dict.keys())
        self.view.pull_down.current(0)

        # gui event bind
        self.view.acquire_btn.config(command=self.launch_popup_window)
        self.view.pull_down.bind("<<ComboboxSelected>>", self.update_microscope_mode)
        self.view.exit_btn.config(command=self.exit_program)

        # framerate information.
        self.framerate = 0

    def progress_bar(self, images_received, microscope_state, mode, stop=False):
        """Update progress bars.

        Parameters
        ----------
        microscope_state : dict
            State of the microscope.  Positions, channels, z-steps, etc.
        mode : str
            Imaging mode.  'live', 'z-stack', ...
        images_received : int
            Number of images received in the controller.
        stop : bool
            Stop flag to set back to 0.

        Examples
        --------
        >>> progress_bar(0, microscope_state, mode)
        """

        if images_received == 0:
            if mode == "live" or mode == "customized":
                # Set to Indeterminate mode.
                self.view.CurAcq["mode"] = "indeterminate"
                self.view.OvrAcq["mode"] = "indeterminate"
            else:
                # Set to Determinate mode and initialize at zero.
                # stack_index = 0
                self.view.CurAcq["mode"] = "determinate"
                self.view.OvrAcq["mode"] = "determinate"
                self.view.CurAcq["value"] = 0
                self.view.OvrAcq["value"] = 0

        # Calculate the number of images anticipated.
        number_of_channels = 0
        for channel in microscope_state["channels"].keys():
            if microscope_state["channels"][channel]["is_selected"] is True:
                number_of_channels += 1

        # Time-lapse acquisition
        number_of_timepoints = int(microscope_state["timepoints"])

        # Multi-Position Acquisition
        if microscope_state["is_multiposition"] is False:
            number_of_positions = 1
        else:
            number_of_positions = len(
                self.parent_controller.configuration["experiment"]["MultiPositions"]
            )

        if mode == "single":
            number_of_slices = 1
        elif mode == "live" or mode == "customized":
            number_of_slices = 1
        elif mode == "z-stack" or "ConstantVelocityAcquisition":
            number_of_slices = microscope_state["number_z_steps"]

        top_anticipated_images = number_of_slices
        bottom_anticipated_images = (
            number_of_channels
            * number_of_slices
            * number_of_timepoints
            * number_of_positions
        )

        if images_received > 0:
            # Update progress bars according to imaging mode.
            if stop is False:
                if mode == "live" or mode == "customized":
                    self.view.CurAcq.start()
                    self.view.OvrAcq.start()
                    self.view.total_acquisition_label.config(text="--:--:--")

                else:
                    # Calculate the number of images remaining.
                    # Time is estimated from the framerate, which includes stage
                    # movement time inherently.
                    try:
                        images_remaining = bottom_anticipated_images - images_received
                        seconds_left = images_remaining / self.framerate
                        self.update_progress_label(seconds_left)
                    except ZeroDivisionError:
                        pass

                if (
                    mode == "z-stack"
                    or mode == "ConstantVelocityAcquisition"
                ):
                    top_percent_complete = 100 * (
                        images_received / top_anticipated_images
                    )

                    self.view.CurAcq["value"] = (
                        top_percent_complete % 100
                        if (top_percent_complete > 100.0)
                        else top_percent_complete
                    )

                    bottom_anticipated_images = 100 * (
                        images_received / bottom_anticipated_images
                    )
                    self.view.OvrAcq["value"] = bottom_anticipated_images

                elif mode == "single":
                    top_percent_complete = 100 * (
                        images_received / top_anticipated_images
                    )
                    self.view.CurAcq["value"] = top_percent_complete
                    self.view.OvrAcq["value"] = top_percent_complete

                else:
                    self.view.CurAcq.start()
                    self.view.OvrAcq.start()

            elif stop is True:
                self.update_progress_label(seconds_left=0)
                self.stop_progress_bar()

    def stop_progress_bar(self):
        """Stop moving the continuous progress bar."""
        self.view.CurAcq.stop()
        self.view.OvrAcq.stop()

    def update_progress_label(self, seconds_left):
        """Update the progress label in the Acquire Bar.

        Formatted time is in HH:MM:SS.

        Parameters
        ----------
        seconds_left : int
            Seconds left in the acquisition.
        """
        hours, remainder = divmod(seconds_left, 3600)
        minutes, seconds = divmod(remainder, 60)
        self.view.total_acquisition_label.config(
            text=f"{int(hours):02}"
                 f":{int(minutes):02}"
                 f":{int(seconds):02}"
        )

    def set_mode(self, mode):
        """Set imaging mode.

        Parameters
        ----------
        mode: str
            Mode could be: 'live', 'z-stack', 'single', 'projection',

        Examples
        --------
        >>> set_mode('live')
        """
        # update pull down combobox
        reverse_dict = dict(map(lambda v: (v[1], v[0]), self.mode_dict.items()))
        if mode not in reverse_dict:
            mode = list(reverse_dict.keys())[0]
        self.mode = mode
        self.view.pull_down.set(reverse_dict[mode])
        self.show_verbose_info("Image mode is set to", mode)

    def get_mode(self):
        """Get the current imaging mode.

        Returns
        -------
        mode : str
            Current imaging mode.

        Examples
        --------
        >>> get_mode()
        """
        return self.mode

    def add_mode(self, mode):
        if mode not in self.mode_dict:
            self.mode_dict[mode] = mode
            self.view.pull_down["values"] = list(self.mode_dict.keys())

    def stop_acquire(self):
        """Stop the acquisition.

        Examples
        --------
        >>> stop_acquire()
        """
        self.stop_progress_bar()
        self.view.acquire_btn.configure(text="Acquire")
        self.view.acquire_btn.configure(state="normal")
        self.view.pull_down.configure(state="normal")
        self.is_acquiring = False

    def set_save_option(self, is_save):
        """Set whether the image will be saved.

        Parameters
        ----------
        is_save : bool
            True if we will save the data.  False if we will not.

        Examples
        --------
        >>> set_save_option(True)
        """
        self.is_save = is_save
        self.parent_controller.configuration["experiment"]["MicroscopeState"][
            "is_save"
        ] = is_save
        self.show_verbose_info("set save data option:", is_save)

    def launch_popup_window(self):
        """Launch the Save Dialog Popup Window

        The popup window should only be launched if the microscope is set to save the
        data, with the exception of the continuous acquisition mode.
        The popup window provides the user with the opportunity to fill in fields that
        describe the experiment and also dictate the save path of the data in a
        standardized format.

        Examples
        --------
        >>> launch_popup_window()
        """
        if self.is_acquiring and self.view.acquire_btn["text"] == "Acquire":
            return
        if not self.is_acquiring and self.view.acquire_btn["text"] == "Stop":
            return
        if self.view.acquire_btn["text"] == "Stop":
            # tell the controller to stop acquire (continuous mode)
            self.view.acquire_btn.configure(state="disabled")
            self.parent_controller.execute("stop_acquire")

        elif self.is_save and self.mode != "live":
            #: object: Instance of the popup save dialog.
            self.acquire_pop = AcquirePopUp(self.view)
            buttons = (
                self.acquire_pop.get_buttons()
            )  # This holds all the buttons in the popup
            widgets = self.acquire_pop.get_widgets()

            # Configure the button callbacks on the popup window
            buttons["Cancel"].config(command=lambda: self.acquire_pop.popup.dismiss())
            buttons["Done"].config(
                command=lambda: self.launch_acquisition(self.acquire_pop)
            )

            # Configure drop down callbacks, will update save settings when file type is
            # changed
            file_type = widgets["file_type"].get_variable()
            file_type.trace_add("write", lambda *args: self.update_file_type(file_type))

            for k, v in self.saving_settings.items():
                if widgets.get(k, None):
                    widgets[k].set(v)

        else:
            self.is_acquiring = True
            self.view.acquire_btn.configure(state="disabled")
            self.view.pull_down.configure(state="disabled")
            self.parent_controller.execute("acquire")

    def update_microscope_mode(self, *args):
        """Gets the state of the pull-down menu and tells the central controller
            Will additionally call functions to disable and enable widgets based on mode

        Parameters

        ----------
        args : str
            Imaging Mode.

        Examples
        --------
        >>> update_microscope_mode('live')
        """
        self.mode = self.mode_dict[self.view.pull_down.get()]
        self.show_verbose_info("The Microscope State is now:", self.get_mode())
        # acquire_bar_controller - update image mode
        self.parent_controller.configuration["experiment"]["MicroscopeState"][
            "image_mode"
        ] = self.mode
        # Update state status of other widgets in the GUI based on what mode is set
        self.parent_controller.channels_tab_controller.set_mode("stop")

    def update_file_type(self, file_type):
        """Updates the file type when the drop down in save dialog is changed.

        Parameters
        ----------
        file_type : str
            File type.

        Examples
        --------
        >>> update_file_type('tiff')
        """
        self.saving_settings["file_type"] = file_type.get()

    def launch_acquisition(self, popup_window):
        """Launch the Acquisition.

        Once the popup window has been filled out, we first create the save path using
        the create_save_path function.
        This automatically removes spaces and replaces them with underscores.
        Then it makes the directory.
        Thereafter, the experiment is ready to go.

        Parameters
        ----------
        popup_window : object
            Instance of the popup save dialog.

        Examples
        --------
        >>> launch_acquisition(popup_window)
        """
        # update saving settings according to user's input
        self.update_experiment_values(popup_window)

        # Verify user's input is non-zero.
        is_valid = (
            self.saving_settings["user"]
            and self.saving_settings["tissue"]
            and self.saving_settings["celltype"]
            and self.saving_settings["label"]
        )

        if is_valid:
            self.is_acquiring = True
            self.view.acquire_btn.configure(state="disabled")
            # Close the window
            popup_window.popup.dismiss()
            # tell central controller, save the image/data
            self.parent_controller.execute("acquire_and_save")

    def exit_program(self):
        """Exit Button

        Quit the software.

        Examples
        --------
        >>> exit_program()
        """
        if messagebox.askyesno("Exit", "Are you sure?"):
            self.show_verbose_info("Exiting Program")
            # call the central controller to stop all the threads
            self.parent_controller.execute("exit")

    def populate_experiment_values(self):
        """Populate the experiment values from the config file.

        Examples
        --------
        >>> populate_experiment_values()
        """
        #: dict: Saving settings.
        self.saving_settings = self.parent_controller.configuration["experiment"][
            "Saving"
        ]
        self.saving_settings["date"] = str(self.saving_settings["date"])
        mode = self.parent_controller.configuration["experiment"]["MicroscopeState"][
            "image_mode"
        ]
        self.set_mode(mode)
        is_save = self.parent_controller.configuration["experiment"]["MicroscopeState"][
            "is_save"
        ]
        self.set_save_option(is_save)

    def update_experiment_values(self, popup_window):
        """Gets the entries from the popup save dialog and overwrites the
        saving_settings dictionary.

        Parameters
        ----------
        popup_window : object
            Instance of the popup save dialog.

        Examples
        --------
        >>> update_experiment_values(popup_window)
        """
        popup_vals = popup_window.get_variables()
        for name in popup_vals:
            # remove leading and tailing whitespaces
            self.saving_settings[name] = popup_vals[name].strip()



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


# Standard library imports
from datetime import datetime
import os
import json
import yaml
from pathlib import Path

# Third party imports

# Local application imports
from navigate.tools.common_functions import copy_proxy_object


def create_save_path(saving_settings):
    """Create path to save the data to.

    This function retrieves the user inputs from the popup save window.
    It then creates a new directory in the user specified path.

    Parameters
    ----------
    saving_settings : dict
        Dictionary containing information necessary to build path to save data to.

    Returns
    -------
    save_directory : str
        Path to save data to.
    """
    root_directory = saving_settings["root_directory"]
    user_string = saving_settings["user"]
    tissue_string = saving_settings["tissue"]
    cell_type_string = saving_settings["celltype"]
    label_string = saving_settings["label"]
    date_string = str(datetime.now().date())

    # Make sure that there are no spaces in the variables
    user_string = user_string.replace(" ", "-")
    tissue_string = tissue_string.replace(" ", "-")
    cell_type_string = cell_type_string.replace(" ", "-")
    label_string = label_string.replace(" ", "-")

    # Create the save directory on disk.
    save_directory = os.path.join(
        root_directory,
        user_string,
        tissue_string,
        cell_type_string,
        label_string,
        date_string,
    )

    os.makedirs(save_directory, exist_ok=True)
    # Determine Number of Cells in Directory
    # Cell1/Position1/1_CH00_000000.tif
    cell_directories = list(
        filter(lambda v: v[:5] == "Cell_", os.listdir(save_directory))
    )
    if len(cell_directories) != 0:
        cell_directories.sort()
        cell_index = int(cell_directories[-1][5:]) + 1
    else:
        cell_index = 1
    cell_string = "Cell_" + str(cell_index).zfill(3)

    save_directory = os.path.join(save_directory, cell_string)
    os.makedirs(save_directory, exist_ok=True)

    # Update the experiment dict
    saving_settings["save_directory"] = save_directory
    saving_settings["date"] = date_string

    return save_directory


def load_yaml_file(file_path):
    """Load YAML file from Disk

    Parameters
    ----------
    file_path : str/os.path
        String or path of the yaml file.

    Returns
    -------
    config_data: dict/list/None
        A dictionary/list of the yaml file content.
        None: if the yaml file has error or not exist.
    """
    file_path = Path(file_path)
    if not file_path.exists():
        return None
    with open(file_path) as f:
        try:
            config_data = yaml.load(f, Loader=yaml.FullLoader)
        except yaml.YAMLError as yaml_error:
            print(f"Can't load yaml file: {file_path} - {yaml_error}")
            return None
    return config_data


def save_yaml_file(file_directory, content_dict, filename="experiment.yml"):
    """Same YAML file to Disk

    Parameters
    ----------
    file_directory : str
        String of directory to save data to.
    content_dict : dict
        Dictionary that holds information about specified content.
    filename : str
        String of name to save data to.  Default is experiment.yml.

    Returns
    -------
    bool
        True if file was saved successfully, False otherwise.
    """
    if file_directory != "" and not os.path.exists(file_directory):
        return False

    try:
        file_name = os.path.join(file_directory, filename)
        if os.path.exists(file_name):
            with open(file_name, "r") as f:
                file_content = f.read()
        else:
            file_content = ""
        with open(file_name, "w") as f:
            f.write(json.dumps(copy_proxy_object(content_dict), indent=4))
    except BaseException:
        with open(file_name, "w") as f:
            f.write(file_content)
        return False
    return True


def delete_folder(top):
    """Delete folder and all sub-folders.

    https://docs.python.org/3/library/os.html#os.walk

    Delete everything reachable from the directory named in "top",
    assuming there are no symbolic links.
    CAUTION:  This is dangerous!  For example, if top == '/', it
    could delete all your disk files.

    Parameters
    ----------
    top : str
        Path to folder to delete.
    """
    for root, dirs, files in os.walk(top, topdown=False):
        for name in files:
            try:
                os.remove(os.path.join(root, name))
            except PermissionError:
                # Windows locks these files sometimes
                pass
        for name in dirs:
            try:
                os.rmdir(os.path.join(root, name))
            except OSError:
                # One of the directories containing a file Windows decided to lock
                pass

    try:
        os.rmdir(top)
    except OSError:
        pass



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#  Standard Imports
import os
import uuid
from pathlib import Path

# Third Party Imports
import tifffile
import numpy.typing as npt

# Local imports
from .data_source import DataSource
from ..metadata_sources.metadata import Metadata
from ..metadata_sources.ome_tiff_metadata import OMETIFFMetadata


class TiffDataSource(DataSource):
    """Data source for TIFF files."""

    def __init__(
        self, file_name: str = "", mode: str = "w", is_bigtiff: bool = False
    ) -> None:
        """Initialize a TIFF data source.

        Parameters
        ----------
        file_name : str
            Path to file.
        mode : str
            File mode. "r" for read, "w" for write.
        is_bigtiff : bool
            Is this a bigtiff file?
        """
        #: np.ndarray: Image data
        self.image = None
        self._write_mode = None
        self._views = []

        super().__init__(file_name, mode)

        #: str: Directory to save the data to.
        self.save_directory = Path(self.file_name).parent

        # Is this an OME-TIFF?
        # TODO: check the header, rather than use the file extension
        if self.file_name.endswith(".ome.tiff") or self.file_name.endswith(".ome.tif"):
            self._is_ome = True
            #: Metadata: Metadata object
            self.metadata = OMETIFFMetadata()
        else:
            self._is_ome = False
            self.metadata = Metadata()

        self._is_bigtiff = is_bigtiff

        # For file writing, do we assume all files end with tiff or tif?
        self.__double_f = self.file_name.endswith("tiff")

        # Keep track of z, time, channel indices
        # self._current_frame = 0
        self._current_time = 0
        self._current_position = 0

    @property
    def data(self) -> npt.ArrayLike:
        """Return the image data as a numpy array.

        Returns
        -------
        npt.ArrayLike
            Image data.
        """
        self.mode = "r"
        return self.image.asarray()

    @property
    def is_bigtiff(self) -> bool:
        """Is this a bigtiff file?

        Returns
        -------
        bool
            Is this a bigtiff file?
        """
        if self._write_mode:
            return self._is_bigtiff
        else:
            return self.image.is_bigtiff

    def set_bigtiff(self, is_bigtiff: bool) -> None:
        """Set whether this is a bigtiff file.

        Parameters
        ----------
        is_bigtiff : bool
            Is this a bigtiff file?
        """
        self._is_bigtiff = is_bigtiff

    @property
    def is_ome(self) -> bool:
        """Is this an OME-TIFF file?

        Returns
        -------
        bool
            Is this an OME-TIFF file?
        """
        if self._write_mode:
            return self._is_ome
        else:
            return self.image.is_ome

    def read(self) -> None:
        """Read a tiff file."""
        self.image = tifffile.TiffFile(self.file_name)

        # TODO: Parse metadata
        for i, ax in enumerate(list(self.image.series[0].axes)):
            if ax == "Q":
                # TODO: This is a hack for tifffile. Find a way to remove this.
                ax = "Z"
            setattr(self, f"shape_{ax.lower()}", self.data.shape[i])

    def write(self, data: npt.ArrayLike, **kw) -> None:
        """Write data to a tiff file.

        One channel, all z-position, one timepoint = one stack.
        N channels are opened simultaneously for writing.
        At each time point, a new file is opened for each channel.

        Parameters
        ----------
        data : npt.ArrayLike
            Data to write to file.
        kw : dict
            Keyword arguments to pass to tifffile.imsave.
        """
        self.mode = "w"

        c, z, self._current_time, self._current_position = self._cztp_indices(
            self._current_frame, self.metadata.per_stack
        )  # find current channel
        if z == 0:
            if c == 0:
                # Make sure we're set up for writing
                self._setup_write_image()
            if self.is_ome:
                ome_xml = self.metadata.to_xml(
                    c=c, t=self._current_time, file_name=self.file_name, uid=self.uid
                ).encode()
        else:
            ome_xml = None

        if len(kw) > 0:
            self._views.append(kw)

        if self.is_ome:
            self.image[c].write(data, description=ome_xml, contiguous=True)
        else:
            dx, dy, dz = self.metadata.voxel_size
            md = {"spacing": dz, "unit": "um", "axes": "ZYX"}
            self.image[c].write(
                data,
                resolution=(1e4 / dx, 1e4 / dy, "CENTIMETER"),
                metadata=md,
                contiguous=True,
            )

        self._current_frame += 1

        # Check if this was the last frame to write
        # print("Switch")
        c, z, _, _ = self._cztp_indices(self._current_frame, self.metadata.per_stack)
        if (z == 0) and (c == 0):
            self.close(True)

    def generate_image_name(self, current_channel, current_time_point):
        """Generates a string for the filename, e.g., CH00_000000.tif

        Parameters
        ----------
        current_channel : int
            Current channel index.
        current_time_point : int
            Current time point index.

        Returns
        -------
        str
            Image name.
        """
        ext = ".ome" if self.is_ome else ""
        ext += ".tiff" if self.__double_f else ".tif"
        image_name = (
            "CH0" + str(current_channel) + "_" + str(current_time_point).zfill(6) + ext
        )
        return image_name

    def _setup_write_image(self) -> None:
        """Setup N channel images for writing."""

        # Grab expected data shape from metadata
        (
            self.shape_x,
            self.shape_y,
            self.shape_c,
            self.shape_z,
            self.shape_t,
        ) = self.metadata.shape
        self.dx, self.dy, self.dz = self.metadata.voxel_size
        self.dc, self.dt = self.metadata.dc, self.metadata.dt

        # Initialize one TIFF per channel per time point
        self.image = []
        self.file_name = []
        self.uid = []
        self._views = []

        if self.metadata._multiposition:
            position_directory = os.path.join(
                self.save_directory, f"Position{self._current_position}"
            )
        else:
            position_directory = self.save_directory
        if not os.path.exists(position_directory):
            os.mkdir(position_directory)
        for ch in range(self.shape_c):
            file_name = os.path.join(
                position_directory, self.generate_image_name(ch, self._current_time)
            )
            self.image.append(
                tifffile.TiffWriter(
                    file_name, bigtiff=self.is_bigtiff, ome=False, byteorder="<"
                )
            )
            self.file_name.append(file_name)
            self.uid.append(str(uuid.uuid4()))

    def _mode_checks(self) -> None:
        """Check that the mode is valid."""
        self._write_mode = self._mode == "w"
        self.close()  # if anything was already open, close it
        if self._write_mode:
            self._current_frame = 0
            self._views = []
            # self._setup_write_image()
        else:
            self.read()
        self._closed = False

    def close(self, internal=False) -> None:
        """Close the file.

        Parameters
        ----------
        internal : bool
            Internal flag. Do not close if True.
        """
        if self._closed and not internal:
            return
        if self.image is None:
            return
        # internal flag needed to avoid _check_shape call until last file is written
        if self._write_mode:
            if not internal:
                self._check_shape(self._current_frame - 1, self.metadata.per_stack)
            for ch in range(len(self.image)):
                self.image[ch].close()
                if self.is_ome and len(self._views) > 0:
                    # Attach OME metadata at the end of the write
                    tifffile.tiffcomment(
                        self.file_name[ch],
                        self.metadata.to_xml(
                            c=ch,
                            t=self._current_time,
                            file_name=self.file_name,
                            uid=self.uid,
                            views=self._views,
                        ).encode(),
                    )
        else:
            self.image.close()
        if not internal:
            self._closed = True



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import tkinter as tk
from tkinter import ttk
import logging

# Third Party Imports
import pandas as pd
from pandastable import Table, Menu, RowHeader, ColumnHeader

# Local Imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class MultiPositionTab(tk.Frame):
    """MultiPositionTab

    MultiPositionTab is a tab in the main window that allows the user to
    create and run multipoint experiments."""

    def __init__(self, setntbk, *args, **kwargs):
        """Initialize the MultiPositionTab

        Parameters
        ----------
        setntbk : ttk.Notebook
            The notebook that contains the settings tab.
        *args : tuple
            Variable length argument list.
        **kwargs : dict
            Arbitrary keyword arguments.
        """
        # Init Frame
        tk.Frame.__init__(self, setntbk, *args, **kwargs)

        #: The index of the tab in the notebook
        self.index = 3

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        #: MultiPointFrame: The frame that contains the widgets for the multipoint
        # experiment settings.
        self.tiling_buttons = MultiPointFrame(self)
        self.tiling_buttons.grid(
            row=0, column=0, columnspan=3, sticky=tk.NSEW, padx=10, pady=10
        )

        #: MultiPointList: The frame that contains the widgets for the multipoint
        # experiment settings.
        self.multipoint_list = MultiPointList(self)
        self.multipoint_list.grid(
            row=6, column=0, columnspan=3, sticky=tk.NSEW, padx=10, pady=10
        )


class MultiPointFrame(ttk.Labelframe):
    """MultiPointFrame

    MultiPointFrame is a frame that contains the widgets for the multipoint
    experiment settings."""

    def __init__(self, settings_tab, *args, **kwargs):
        """Initialize the MultiPointFrame

        Parameters
        ----------
        settings_tab : tk.Frame
            The frame that contains the settings tab.
        *args : tuple
            Variable length argument list.
        **kwargs : dict
            Arbitrary keyword arguments.

        """
        text_label = "Multi-Position Acquisition"
        ttk.Labelframe.__init__(self, settings_tab, text=text_label, *args, **kwargs)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        #: dict: A dictionary of all the widgets that are tied to each widget name.
        self.buttons = {
            "tiling": ttk.Button(self, text="Launch Tiling Wizard"),
            "save_data": ttk.Button(self, text="Save Positions to Disk"),
            "load_data": ttk.Button(self, text="Load Positions from Disk"),
            "eliminate_tiles": ttk.Button(self, text="Eliminate Empty Positions"),
        }
        counter = 0
        for key, button in self.buttons.items():
            if counter == 0:
                row, column = 0, 0
            elif counter == 1:
                row, column = 1, 0
            elif counter == 2:
                row, column = 1, 1
            elif counter == 3:
                row, column = 0, 1

            button.grid(
                row=row, column=column, sticky=tk.NSEW, padx=(4, 1), pady=(4, 6)
            )
            counter += 1

    def get_variables(self):
        """Returns a dictionary of all the variables that are tied to each widget name.

        The key is the widget name, value is the variable associated.

        Returns
        -------
        dict
            A dictionary of all the variables that are tied to each widget name.
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get_variable()
        return variables

    def get_widgets(self):
        """Returns a dictionary of all the widgets that are tied to each widget name.

        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        dict
            A dictionary of all the widgets that are tied to each widget name.
        """
        return self.inputs


class MultiPointList(ttk.Frame):
    """MultiPointList

    MultiPointList is a frame that contains the widgets for the multipoint
    experiment settings. uses Pandastable for embedding an interactive list within a
    tk Frame. https://pandastable.readthedocs.io/en/latest/
    """

    def __init__(self, settings_tab, *args, **kwargs):
        """Initialize the MultiPointList

        Parameters
        ----------
        settings_tab : tk.Frame
            The frame that contains the settings tab.
        *args : tuple
            Variable length argument list.
        **kwargs : dict
            Arbitrary keyword arguments.
        """
        ttk.Frame.__init__(self, settings_tab, *args, **kwargs)

        df = pd.DataFrame({"X": [0], "Y": [0], "Z": [0], "R": [0], "F": [0]})
        #: MultiPositionTable: The PandasTable instance that is being used.
        self.pt = MultiPositionTable(self, showtoolbar=False)
        self.pt.show()
        self.pt.model.df = df

    def get_table(self):
        """Returns a reference to multipoint table dataframe.

        Parameters
        ----------
        self : object
            Multipoint List instance

        Returns
        -------
        self.pt.model.df: Pandas DataFrame
            Reference to table data as dataframe
        """
        return self.pt


class MultiPositionRowHeader(RowHeader):
    """MultiPositionRowHeader

    MultiPositionRowHeader is a class that inherits from RowHeader. It is used to
    customize the row header for the multipoint table.
    """

    def __init__(self, parent=None, table=None, width=50):
        """Initialize the MultiPositionRowHeader

        Parameters
        ----------
        parent : tk.Frame
            The frame that contains the settings tab.
        table : PandasTable
            The PandasTable instance that is being used.
        width : int
            The width of the row header.
        """
        super().__init__(parent, table, width)

    def popupMenu(self, event, rows=None, cols=None, outside=None):
        """Add right click behaviour for row header

        Parameters
        ----------
        event : tk.Event
            The event that triggers the popup menu.
        rows : list
            The list of rows that are selected.
        cols : list
            The list of columns that are selected.
        outside : bool
            Whether the popup menu is triggered outside the table.

        Returns
        -------
        popupmenu : tk.Menu
            The popup menu.
        """

        defaultactions = {
            "Sort by index": lambda: self.table.sortTable(index=True),
            "Reset index": lambda: self.table.resetIndex(),
            "Toggle index": lambda: self.toggleIndex(),
            "Copy index to column": lambda: self.table.copyIndex(),
            "Rename index": lambda: self.table.renameIndex(),
            "Sort columns by row": lambda: self.table.sortColumnIndex(),
            "Select All": self.table.selectAll,
            "Insert New Position": self.table.insertRow,
            "Add Current Position": self.table.addStagePosition,
            "Add New Position(s)": lambda: self.table.addRows(),
            "Delete Position(s)": lambda: self.table.deleteRow(),
            "Duplicate Row(s)": lambda: self.table.duplicateRows(),
            "Set Row Color": lambda: self.table.setRowColors(cols="all"),
        }
        main = [
            "Insert New Position",
            "Add Current Position",
            "Add New Position(s)",
            "Delete Position(s)",
        ]

        popupmenu = Menu(self, tearoff=0)

        def popupFocusOut(event):
            popupmenu.unpost()

        for action in main:
            popupmenu.add_command(label=action, command=defaultactions[action])

        popupmenu.bind("<FocusOut>", popupFocusOut)
        popupmenu.focus_set()
        popupmenu.post(event.x_root, event.y_root)
        # applyStyle(popupmenu)
        return popupmenu


class MultiPositionColumnHeader(ColumnHeader):
    """MultiPositionColumnHeader

    MultiPositionColumnHeader is a class that inherits from ColumnHeader. It is used to
    customize the column header for the multipoint table.
    """

    def __init__(self, parent=None, table=None, bg="gray25"):
        """Initialize the MultiPositionColumnHeader

        Parameters
        ----------
        parent : tk.Frame
            The frame that contains the settings tab.
        table : PandasTable
            The PandasTable instance that is being used.
        bg : str
            The background color of the column header.
        """

        super().__init__(parent, table, bg)

    def popupMenu(self, event):
        """Add left and right click behaviour for column header

        Parameters
        ----------
        event : tk.Event
            The event that triggers the popup menu.

        Returns
        -------
        popupmenu : tk.Menu
            The popup menu.
        """

        df = self.table.model.df
        if len(df.columns) == 0:
            return

        multicols = self.table.multiplecollist
        colnames = list(df.columns[multicols])[:4]
        colnames = [str(i)[:20] for i in colnames]
        if len(colnames) > 2:
            colnames = ",".join(colnames[:2]) + "+%s others" % str(len(colnames) - 2)
        else:
            colnames = ",".join(colnames)
        popupmenu = Menu(self, tearoff=0)

        def popupFocusOut(event):
            """Unpost the popup menu"""
            popupmenu.unpost()

        popupmenu.add_command(
            label="Sort by " + colnames + " \u2193",
            command=lambda: self.table.sortTable(
                columnIndex=multicols, ascending=[0 for i in multicols]
            ),
        )
        popupmenu.add_command(
            label="Sort by " + colnames + " \u2191",
            command=lambda: self.table.sortTable(
                columnIndex=multicols, ascending=[1 for i in multicols]
            ),
        )
        popupmenu.bind("<FocusOut>", popupFocusOut)
        popupmenu.focus_set()
        popupmenu.post(event.x_root, event.y_root)
        # applyStyle(popupmenu)
        return popupmenu


class MultiPositionTable(Table):
    """MultiPositionTable

    MultiPositionTable is a class that inherits from Table. It is used to
    customize the table for the multipoint table.
    """

    def __init__(self, parent=None, **kwargs):
        """Initialize the MultiPositionTable

        Parameters
        ----------
        parent : tk.Frame
            The frame that contains the settings tab.
        **kwargs : dict
            Arbitrary keyword arguments.
        """

        super().__init__(parent, width=400, height=500, columns=4, **kwargs)

        self.loadCSV = None
        self.exportCSV = None
        self.insertRow = None
        self.addStagePosition = None

    def show(self, callback=None):
        """Show the table

        Parameters
        ----------
        callback : function
            The function that is called when the table is shown.
        """
        super().show(callback)

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        #: MultiPositionRowHeader: The row header for the table.
        self.rowheader = MultiPositionRowHeader(self.parentframe, self)
        self.rowheader.grid(row=1, column=0, rowspan=1, sticky="news")

        #: MultiPositionColumnHeader: The column header for the table.
        self.tablecolheader = MultiPositionColumnHeader(
            self.parentframe, self, bg=self.colheadercolor
        )
        self.tablecolheader.grid(row=0, column=1, rowspan=1, sticky="news")

    def popupMenu(self, event, rows=None, cols=None, outside=None):
        """Add right click behaviour for table

        Parameters
        ----------
        event : tk.Event
            The event that triggers the popup menu.
        rows : list
            The list of rows that are selected.
        cols : list
            The list of columns that are selected.
        outside : bool
            Whether the popup menu is triggered outside the table.

        Returns
        -------
        popupmenu : tk.Menu
            The popup menu.
        """
        popupmenu = Menu(self, tearoff=0)

        def popupFocusOut(event):
            popupmenu.unpost()

        popupmenu.add_command(label="Load Positions from Disk", command=self.loadCSV)
        popupmenu.add_command(label="Save Positions to Disk", command=self.exportCSV)
        popupmenu.bind("<FocusOut>", popupFocusOut)
        popupmenu.focus_set()
        popupmenu.post(event.x_root, event.y_root)
        return popupmenu



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import tkinter as tk
from tkinter import ttk
import logging

# Third Party Imports
from matplotlib.figure import Figure
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import matplotlib.ticker as tck

# Local Imports
from navigate.view.custom_widgets.popup import PopUp
from navigate.view.custom_widgets.LabelInputWidgetFactory import LabelInput
from navigate.view.custom_widgets.validation import ValidatedSpinbox

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class AutofocusPopup:
    """Class creates the popup to configure autofocus parameters."""

    def __init__(self, root, *args, **kwargs):
        """Initialize the AutofocusPopup class.

        Parameters
        ----------
        root : tk.Tk
            Root window.
        *args
            Variable length argument list.
        **kwargs
            Arbitrary keyword arguments.
        """
        # Creating popup window with this name and size/placement,
        # PopUp is a Toplevel window
        #: PopUp: PopUp window
        self.popup = PopUp(
            root, "Autofocus Settings", "+320+180", top=False, transient=False
        )
        # Change background of popup window to white
        self.popup.configure(bg="white")

        # Creating content frame
        content_frame = self.popup.get_frame()
        device_frame = tk.Frame(content_frame)
        device_frame.grid(row=0, columnspan=3, sticky=tk.NSEW)

        # Dictionary for all the variables
        #: dict: Dictionary of all the input widgets.
        self.inputs = {}
        #: dict: Dictionary of all the setting variables.
        self.setting_vars = {}

        self.inputs["device"] = LabelInput(
            parent=device_frame,
            label="Device Type",
            input_class=ttk.Combobox,
            input_var=tk.StringVar(),
            input_args={"width": 20, "state": "readonly"},
            label_args={"padding": (0, 0, 10, 0)},
        )
        self.inputs["device"].grid(row=0, column=0, pady=3, padx=5)

        self.inputs["device_ref"] = LabelInput(
            parent=device_frame,
            label="Device Reference",
            input_class=ttk.Combobox,
            input_var=tk.StringVar(),
            input_args={"width": 20, "state": "readonly"},
            label_args={"padding": (0, 0, 10, 0)},
        )
        self.inputs["device_ref"].grid(row=0, column=1, pady=3, padx=30)

        starting_row_id = 1

        # Row 0, Column Titles
        title_labels = ["Select", "Range", "Step Size"]
        for i in range(3):
            title = ttk.Label(content_frame, text=title_labels[i], padding=(2, 5, 0, 0))
            title.grid(row=starting_row_id, column=i, sticky=tk.NSEW)

        # Row 1, 2 - Autofocus Settings
        setting_names = ["coarse", "fine", "robust_fit"]
        setting_labels = ["Coarse", "Fine", "Inverse Power Tent Fit"]
        for i in range(2):
            # Column 0 - Checkboxes
            variable = tk.BooleanVar(False)
            widget = ttk.Checkbutton(
                content_frame, text=setting_labels[i], variable=variable
            )
            widget.grid(row=i + 1 + starting_row_id, column=0, sticky=tk.NSEW, padx=5)
            self.setting_vars[setting_names[i] + "_selected"] = variable

            # Column 1 - Ranges
            widget = LabelInput(
                parent=content_frame,
                input_class=ValidatedSpinbox,
                input_var=tk.StringVar(),
                input_args={"from_": 0.0, "to": 50000},
            )
            widget.grid(
                row=i + 1 + starting_row_id,
                column=1,
                sticky=tk.NSEW,
                padx=(0, 5),
                pady=(15, 0),
            )
            self.inputs[setting_names[i] + "_range"] = widget
            self.setting_vars[setting_names[i] + "_range"] = widget.get_variable()

            # Column 2 - Step Sizes
            widget = LabelInput(
                parent=content_frame,
                input_class=ValidatedSpinbox,
                input_var=tk.StringVar(),
                input_args={"from_": 0.0, "to": 50000},
            )
            widget.grid(
                row=i + 1 + starting_row_id,
                column=2,
                sticky=tk.NSEW,
                padx=(0, 5),
                pady=(15, 0),
            )
            self.inputs[setting_names[i] + "_step_size"] = widget
            self.setting_vars[setting_names[i] + "_step_size"] = widget.get_variable()

        # Row 4, Autofocus Button
        #: ttk.Button: Autofocus button.
        self.autofocus_btn = ttk.Button(content_frame, text="Autofocus")
        self.autofocus_btn.grid(
            row=starting_row_id + 4, column=2, padx=(0, 25), pady=(10, 10), sticky=tk.E
        )

        variable = tk.BooleanVar(False)
        robust_fit = ttk.Checkbutton(
            content_frame, text=setting_labels[2], variable=variable
        )
        robust_fit.grid(row=starting_row_id + 4, column=0, sticky=tk.NSEW, padx=5)
        self.setting_vars["robust_fit"] = variable

        # Row 5, Plot
        #: matplotlib.figure.Figure: Figure for the plot.
        self.fig = Figure(figsize=(5, 5), dpi=100)
        #: matplotlib.axes.Axes: Axes for the plot.
        self.coarse = self.fig.add_subplot(111)
        self.coarse.set_title("Discrete Cosine Transform", fontsize=18)
        self.coarse.set_xlabel("Focus Stage Position", fontsize=16)
        self.coarse.yaxis.set_minor_locator(tck.AutoMinorLocator())
        self.coarse.xaxis.set_minor_locator(tck.AutoMinorLocator())

        self.fig.tight_layout()
        canvas = FigureCanvasTkAgg(self.fig, master=content_frame)
        canvas.draw()
        canvas.get_tk_widget().grid(
            row=starting_row_id + 5,
            column=0,
            columnspan=3,
            sticky=tk.NSEW,
            padx=(5, 5),
            pady=(5, 5),
        )

    def get_widgets(self):
        """Returns the dictionary of input widgets.

        Returns
        -------
        dict
            Dictionary of all the input widgets.
        """
        return self.inputs



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard library imports
import tkinter as tk
from tkinter import ttk
import logging
import platform

# Third party imports

# Local application imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class DockableNotebook(ttk.Notebook):
    """Dockable Notebook that allows for tabs to be popped out into a separate
    windows by right clicking on the tab. The tab must be selected before
    right-clicking.
    """

    def __init__(self, parent, root, *args, **kwargs):
        """Initialize Dockable Notebook

        Parameters
        ----------
        parent: Tk parent widget.
            The parent widget being passed down for hierarchy and organization.
            Typically a ttk.Frame or tk.Frame.
        root : Tk top-level widget.
            Tk.tk GUI instance.
        *args :
            Options for the ttk.Notebook class
        **kwargs:
            Keyword options for the ttk.Notebook class
        """
        ttk.Notebook.__init__(self, parent, *args, **kwargs)
        #: tk.Tk: Tkinter root
        self.root = root
        #: list: List of tab variables
        self.tab_list = []

        # Formatting
        tk.Grid.columnconfigure(self, "all", weight=1)
        tk.Grid.rowconfigure(self, "all", weight=1)

        # Popup setup
        #: tk.Menu: Tkinter menu
        self.menu = tk.Menu(self, tearoff=0)
        self.menu.add_command(label="Popout Tab", command=self.popout)

        # Bindings
        if platform.system() == "Darwin":
            self.bind("<ButtonPress-2>", self.find)
        else:
            self.bind("<ButtonPress-3>", self.find)

    def set_tablist(self, tab_list):
        """Setter for tab list

        Parameters
        ----------
        tab_list: list
            List of tab variables
        """
        self.tab_list = tab_list

    def get_absolute_position(self):
        """Get absolute position of mouse.

        This helps the popup menu appear where the mouse is right clicked.

        Returns
        -------
        x, y: integers
            Coordinates to be used (x, y)
        """
        x = self.root.winfo_pointerx()
        y = self.root.winfo_pointery()
        return x, y

    def find(self, event):
        """Find the widget that was clicked on.

        Will check if the proper widget element in the event is what we expect.

        In this case its checking that the label in the tab has been selected.
        It then gets the proper position and calls the popup.

        Parameters
        ----------
        event: Tkinter event
            Holds information about the event that was triggered and caught by Tkinters
            event system
        """
        element = event.widget.identify(event.x, event.y)
        if "label" in element:
            try:
                x, y = self.get_absolute_position()
                self.menu.tk_popup(x, y)
            finally:
                self.menu.grab_release()

    def popout(self):
        """Popout the currently selected tab.

        Gets the currently selected tab, the tabs name and checks if the tab name is
        in the tab list. If the tab is in the list, its removed from the list,
        hidden, and then passed to a new Top Level window.
        """
        # Get ref to correct tab to popout
        tab = self.select()
        tab_text = self.tab(tab)["text"]
        for tab_name in self.tab_list:
            if tab_text == self.tab(tab_name)["text"]:
                tab = tab_name
                self.tab_list.remove(tab_name)
        self.hide(tab)
        self.root.wm_manage(tab)

        # self.root.wm_title(tab, tab_text)
        tk.Wm.title(tab, tab_text)
        tk.Wm.protocol(tab, "WM_DELETE_WINDOW", lambda: self.dismiss(tab, tab_text))
        if tab_text == "Camera View":
            tk.Wm.minsize(tab, 663, 597)
            tab.is_docked = False
        elif tab_text == "Waveform Settings":
            tab.is_docked = False

    def dismiss(self, tab, tab_text):
        """Dismisses the popup menu

        This function is called when the top level that the tab was originally passed to
        has been closed. The window manager releases control and then the tab is
        added back to its original ttk.Notebook.

        Parameters
        ----------
        tab: Tkinter tab (path to widget represented as a str)
            The tab that was popped out, this reference to the dismiss function is
            associated with this tab
        tab_text: string
            Name of the tab as it appears in the GUI
        """
        self.root.wm_forget(tab)
        tab.grid(row=0, column=0)
        if self.index("end") - 1 > tab.index:
            self.insert(tab.index, tab)
        else:
            self.insert("end", tab)
        self.tab(tab, text=tab_text)
        self.tab_list.append(tab)
        if tab_text == "Camera View":
            tab.canvas.configure(width=512, height=512)
            tab.is_docked = True
        elif tab_text == "Waveform Settings":
            tab.is_docked = True



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

#  Standard Imports
import logging

# Third Party Imports
import numpy as np
from scipy.fftpack import dctn

# Local Imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


def fast_normalized_dct_shannon_entropy(input_array, psf_support_diameter_xy):
    """Calculates the entropy of an image.

    Parameters
    ----------
    input_array : np.ndarray
        2D or 3D image.  If 3D, will iterate through each 2D plane.
    psf_support_diameter_xy : float
        Support for the PSF in the x-dimension.

    Returns
    -------
    entropy : np.ndarray
        Entropy value.
    """

    dct_array = dctn(input_array, type=2)
    abs_array = np.abs(dct_array / np.linalg.norm(dct_array))
    yh = int(input_array.shape[1] // psf_support_diameter_xy)
    xh = int(input_array.shape[0] // psf_support_diameter_xy)
    entropy = (
        -2 * np.nansum(abs_array[:xh, :yh] * np.log2(abs_array[:xh, :yh])) / (yh * xh)
    )

    return np.atleast_1d(entropy)



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import base64
import requests
import numpy
import json
from io import BytesIO
from math import ceil

import logging

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


def prepare_service(service_url, **kwargs):
    """Prepare service for Ilastik segmentation

    Parameters
    ----------
    service_url : str
        url of the service
    kwargs : dict
        project_file : str
        project file for Ilastik segmentation

    Returns
    -------
    dict
        response from the server
    """
    service_url = service_url.rstrip("/")
    if service_url.endswith("ilastik"):
        r = requests.get(f"{service_url}/load?project={kwargs['project_file']}")
        logger.info(f"get response:{r.status_code}, {r.content}")
        if r.status_code == 200:
            return json.loads(r.content)
    return None


class IlastikSegmentation:
    """Ilastik segmentation class.

    Uses Ilastik REST API to perform segmentation in a separate process.
    """

    def __init__(self, model):
        """Initialize Ilastik segmentation class.

        Parameters
        ----------
        model : Model
            Model object
        """

        #: navigate.model.Model: Model object
        self.model = model

        #: str: url of the service
        self.service_url = self.model.configuration["rest_api_config"]["Ilastik"]["url"]

        #: str: project file for Ilastik segmentation
        self.project_file = None

        #: str: resolution of the current microscope
        self.resolution = None

        #: str: zoom of the current microscope
        self.zoom = None

        #: int: number of pieces
        self.pieces_num = 1

        #: int: size of pieces
        self.pieces_size = 1

        #: dict: configuration table
        self.config_table = {"data": {"init": self.init_func, "main": self.data_func}}

    def init_func(self, *args):
        """Initialize Ilastik segmentation.

        Parameters
        ----------
        args : list
            list of arguments
        """
        if (
            self.resolution
            != self.model.configuration["experiment"]["MicroscopeState"][
                "microscope_name"
            ]
            or self.zoom
            != self.model.configuration["experiment"]["MicroscopeState"]["zoom"]
        ):
            self.update_setting()

    def data_func(self, frame_ids):
        """Perform Ilastik segmentation.

        Parameters
        ----------
        frame_ids : list
            list of frame ids
        """
        # Ilastik process multiple images in sequence.
        img_data = [base64.b64encode(self.model.data_buffer[idx]) for idx in frame_ids]
        json_data = {
            "dtype": "uint16",
            "shape": (self.model.img_height, self.model.img_width),
            "image": [img.decode("utf-8") for img in img_data],
        }

        response = requests.post(
            f"{self.service_url}/segmentation", json=json_data, stream=True
        )
        if response.status_code == 200:
            # segmentation_mask is a dictionary like object with keys 'arr_0',
            # 'arr_1'...
            segmentation_mask = numpy.load(BytesIO(response.raw.read()))
            # display segmentation
            for idx in range(len(segmentation_mask)):
                segmentation_id = "arr_{}".format(idx)
                if self.model.display_ilastik_segmentation:
                    self.model.event_queue.put(
                        ("ilastik_mask", segmentation_mask[segmentation_id])
                    )
                # mark position
                if self.model.mark_ilastik_position:
                    self.mark_position(segmentation_mask[segmentation_id])
        else:
            print("There is something wrong!")

    def update_setting(self):
        """Update Ilastik segmentation settings."""
        self.resolution = self.model.configuration["experiment"]["MicroscopeState"][
            "microscope_name"
        ]
        self.zoom = self.model.configuration["experiment"]["MicroscopeState"]["zoom"]
        # Get current mag
        current_microscope_name = self.resolution
        curr_pixel_size = float(
            self.model.configuration["configuration"]["microscopes"][
                current_microscope_name
            ]["zoom"]["pixel_size"][self.zoom]
        )
        # target resolution is 'high'
        # TODO:
        high_res_microscope_name = "Nanoscale"
        pixel_size = float(
            self.model.configuration["configuration"]["microscopes"][
                high_res_microscope_name
            ]["zoom"]["pixel_size"]["N/A"]
        )
        # calculate pieces
        self.pieces_num = int(curr_pixel_size / pixel_size)
        self.pieces_size = ceil(
            float(
                self.model.configuration["experiment"]["CameraParameters"]["x_pixels"]
            )
            / self.pieces_num
        )

        #: float: position step size
        self.posistion_step_size = self.pieces_size * pixel_size
        # calculate corner (x,y)
        curr_fov_x = (
            float(
                self.model.configuration["experiment"]["CameraParameters"]["x_pixels"]
            )
            * curr_pixel_size
        )
        curr_fov_y = (
            float(
                self.model.configuration["experiment"]["CameraParameters"]["y_pixels"]
            )
            * curr_pixel_size
        )

        #: float: x start position
        self.x_start = (
            float(self.model.configuration["experiment"]["StageParameters"]["x"])
            - curr_fov_x / 2
        )

        #: float: y start position
        self.y_start = (
            float(self.model.configuration["experiment"]["StageParameters"]["y"])
            - curr_fov_y / 2
        )

    def mark_position(self, mask):
        """Mark position based on the segmentation mask.

        Parameters
        ----------
        mask : numpy.ndarray
            segmentation mask
        """

        # target_label = self.model.ilastik_target
        target_label = self.model.ilastik_target_labels
        lx, rx = 0, self.pieces_size
        # get current z, theta, focus
        # TODO: are they same as high resolution?
        z = self.model.configuration["experiment"]["StageParameters"]["z"]
        theta = self.model.configuration["experiment"]["StageParameters"]["theta"]
        f = self.model.configuration["experiment"]["StageParameters"]["f"]
        pos_x, pos_y = self.x_start, self.y_start
        table_values = []
        for i in range(self.pieces_num):
            ly, ry = 0, self.pieces_size
            for j in range(self.pieces_num):
                for k in target_label:
                    if numpy.any(mask[lx:rx, ly:ry, 0] == k):
                        table_values.append([pos_x, pos_y, z, theta, f])
                        break
                pos_y += self.posistion_step_size
                ly += self.pieces_size
                ry += self.pieces_size
            lx += self.pieces_size
            rx += self.pieces_size
            pos_x += self.posistion_step_size
            pos_y = self.y_start
        self.model.event_queue.put(("multiposition", table_values))



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#
#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#
#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.
#
# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Imports
import logging
from multiprocessing.managers import ListProxy

# Third Party Imports

# Local Imports

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class StageBase:
    """Stage Parent Class"""

    def __init__(self, microscope_name, device_connection, configuration, device_id=0):
        """Initialize the stage.

        Parameters
        ----------
        microscope_name : str
            Name of microscope in configuration
        device_connection : object
            Hardware device to connect to
        configuration : multiprocessing.managers.DictProxy
            Global configuration of the microscope
        device_id : int, optional
            Device ID, by default 0
        """
        stage_configuration = configuration["configuration"]["microscopes"][
            microscope_name
        ]["stage"]
        if type(stage_configuration["hardware"]) == ListProxy:

            #: list: List of stage axes available.
            self.axes = list(stage_configuration["hardware"][device_id]["axes"])

            device_axes = stage_configuration["hardware"][device_id].get(
                "axes_mapping", []
            )

            #: int: Feedback alignment value for the stage. Default is None.
            self.stage_feedback = stage_configuration["hardware"][device_id].get(
                "feedback_alignment", None
            )
        else:
            self.axes = list(stage_configuration["hardware"]["axes"])
            device_axes = stage_configuration["hardware"].get("axes_mapping", [])
            self.stage_feedback = stage_configuration["hardware"].get(
                "feedback_alignment", None
            )

        if device_axes is None:
            device_axes = []

        if len(self.axes) > len(device_axes):
            log_string = (
                f"{microscope_name}: Stage axes mapping is not specified in "
                "the configuration file, will use default settings!"
            )
            logger.debug(log_string)
            print(log_string)

        #: dict: Dictionary of stage axes and their corresponding hardware axes.
        self.axes_mapping = dict(zip(self.axes, device_axes))

        """Initial setting for all positions
        self.x_pos, self.y_pos etc are the true axis positions, no matter whether
        the stages are zeroed or not. All units are in microns.
        """
        for ax in self.axes:
            setattr(self, f"{ax}_pos", 0)
            setattr(self, f"{ax}_min", stage_configuration[f"{ax}_min"])
            setattr(self, f"{ax}_max", stage_configuration[f"{ax}_max"])

        #: bool: Whether the stage has limits enabled or not. Default is True.
        self.stage_limits = True

    def get_position_dict(self):
        """Return a dictionary with the saved stage positions.

        Returns
        -------
        dict
            Dictionary of stage positions.
        """
        position_dict = {}
        for ax in self.axes:
            ax_str = f"{ax}_pos"
            position_dict[ax_str] = getattr(self, ax_str)
        return position_dict

    def get_abs_position(self, axis, axis_abs):
        """Ensure the requested position is within axis bounds and return it.

        Parameters
        ----------
        axis : str
            An axis: x, y, z, f, theta
        axis_abs : float
            Absolute position value

        Returns
        -------
        float
            Position to move the stage to for this axis.
        """
        try:
            # Get all necessary attributes.
            # If we can't we'll move to the error case (e.g., -1e50).
            axis_min, axis_max = getattr(self, f"{axis}_min"), getattr(
                self, f"{axis}_max"
            )
        except (KeyError, AttributeError) as e:
            # Alert the user, but don't kill the thread
            msg = f"No key {e} in move_dictionary or axis missing from {self.axes}."
            logger.debug(msg)
            print(msg)
            return -1e50

        if not self.stage_limits:
            return axis_abs

        # Check that our position is within the axis bounds, fail if it's not.
        if (axis_min > axis_abs) or (axis_max < axis_abs):
            log_string = (
                f"Absolute movement stopped: {axis} limit would be reached!"
                f"{axis_abs} is not in the range {axis_min} to {axis_max}."
            )
            logger.info(log_string)
            print(log_string)
            # Return a ridiculous value to make it clear we've failed.
            # This is to avoid returning a duck type.
            return -1e50
        return axis_abs

    def verify_abs_position(self, move_dictionary, is_strict=False):
        """Ensure the requested moving positions are within axes bounds

        Parameters
        ----------
        move_dictionary : dict
            A dictionary of values required for movement.
            Includes 'x_abs', 'y_abs', etc. for one or more axes.
            Expect values in micrometers, except for theta, which is in degrees.
        is_strict : bool, optional
            If True, return an empty dict if any axis is out of bounds.

        Returns
        -------
        dict
            a verified moving dict {axis: abs_position}
        """
        abs_pos_dict = {}
        result_flag = True
        for axis in self.axes_mapping.keys():
            if f"{axis}_abs" not in move_dictionary:
                continue
            axis_abs = move_dictionary[f"{axis}_abs"]
            axis_min = getattr(self, f"{axis}_min")
            axis_max = getattr(self, f"{axis}_max")

            if self.stage_limits and ((axis_abs < axis_min) or (axis_abs > axis_max)):
                log_string = (
                    f"Absolute movement stopped: {axis} limit would be reached!"
                    f"{axis_abs} is not in the range {axis_min} to {axis_max}."
                )
                logger.info(log_string)
                print(log_string)
                result_flag = False
            else:
                abs_pos_dict[axis] = axis_abs

        if is_strict and not result_flag:
            return {}
        return abs_pos_dict

    def stop(self):
        """Stop all stage movement abruptly."""
        pass

    def close(self):
        """Close the stage."""
        pass



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import logging

# Third Party Imports

# Local Imports
from navigate.controller.sub_controllers.gui_controller import GUIController

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class CameraSettingController(GUIController):
    """Controller for the camera settings."""

    def __init__(self, view, parent_controller=None):
        """Initialize the camera setting controller.

        Parameters
        ----------
        view : tkinter.Frame
            The view for the camera settings.
        parent_controller : navigate.controller.main_controller.MainController
            The parent controller.
        """
        super().__init__(view, parent_controller)

        # default values
        #: bool: True if in initialization
        self.in_initialization = True
        #: str: Resolution value
        self.resolution_value = "1x"
        #: str: Mode value
        self.mode = "stop"
        #: str: Solvent type
        self.solvent = "BABB"

        # Getting Widgets/Buttons
        #: dict: Mode widgets
        self.mode_widgets = view.camera_mode.get_widgets()
        #: dict: Framerate widgets
        self.framerate_widgets = view.framerate_info.get_widgets()
        #: dict: ROI widgets
        self.roi_widgets = view.camera_roi.get_widgets()
        #: dict: ROI buttons
        self.roi_btns = view.camera_roi.get_buttons()

        # initialize
        #: int: Default pixel size
        self.default_pixel_size = None
        #: int: Default width
        #: int: Default height
        self.default_width, self.default_height = None, None
        #: int: Trigger type - 1: Internal, 2: External, 3: Synchronous
        self.trigger_source = None
        #: int: Trigger active - 1: Edge, 2: Level, 3: Synchronous
        self.trigger_active = None
        #: int: Readout speed
        self.readout_speed = None
        #: int: Camera width step interval
        self.step_width = 4
        #: int: Camera height step interval
        self.step_height = 4
        #: int: Camera width minimum
        self.min_width = 4
        #: int: Camera height minimum
        self.min_height = 4
        self.initialize()

        # Event binding
        #: bool: True if pixel event id
        self.pixel_event_id = None
        self.mode_widgets["Sensor"].widget.bind(
            "<<ComboboxSelected>>", self.update_sensor_mode
        )
        self.mode_widgets["Pixels"].get_variable().trace_add(
            "write", self.update_number_of_pixels
        )
        self.roi_widgets["Width"].get_variable().trace_add("write", self.update_fov)
        self.roi_widgets["Height"].get_variable().trace_add("write", self.update_fov)

        for btn_name in self.roi_btns:
            self.roi_btns[btn_name].config(command=self.update_roi(btn_name))

    def initialize(self):
        """Sets widgets based on data given from main controller/config."""

        # Get Default Configuration Values
        camera_config_dict = (
            self.parent_controller.configuration_controller.camera_config_dict
        )
        if camera_config_dict is None:
            return

        self.update_camera_device_related_setting()

        # Camera Mode
        self.mode_widgets["Sensor"].widget["values"] = ["Normal", "Light-Sheet"]
        self.mode_widgets["Sensor"].widget["state"] = "readonly"
        self.mode_widgets["Sensor"].widget.selection_clear()

        # Readout Mode
        self.camera_readout_directions = camera_config_dict[
            "supported_readout_directions"
        ]
        self.mode_widgets["Readout"].widget["values"] = self.camera_readout_directions

        self.mode_widgets["Readout"].widget["state"] = "disabled"
        self.mode_widgets["Readout"].selection_clear()

        # Pixels
        self.mode_widgets["Pixels"].widget["state"] = "disabled"
        self.mode_widgets["Pixels"].set("")
        self.mode_widgets["Pixels"].widget.config(from_=1)  # min value
        self.mode_widgets["Pixels"].widget.config(
            to=self.default_height / 2
        )  # max value
        self.mode_widgets["Pixels"].widget.config(increment=1)  # step value

        self.framerate_widgets["exposure_time"].widget["state"] = "disabled"
        self.framerate_widgets["readout_time"].widget["state"] = "disabled"
        self.framerate_widgets["max_framerate"].widget["state"] = "disabled"

        # Set range value
        self.roi_widgets["Width"].widget.config(from_=2)
        self.roi_widgets["Width"].widget.config(increment=2)
        self.roi_widgets["Height"].widget.config(from_=2)
        self.roi_widgets["Height"].widget.config(increment=2)

        # set binning options
        self.roi_widgets["Binning"].widget["values"] = [
            "{}x{}".format(i, i) for i in [1, 2, 4]
        ]
        self.roi_widgets["Binning"].widget["state"] = "readonly"

        # This should not be edited for now
        # Center position
        self.roi_widgets["Center_X"].widget["state"] = "disabled"
        self.roi_widgets["Center_Y"].widget["state"] = "disabled"

        # FOV
        self.roi_widgets["FOV_X"].widget["state"] = "disabled"
        self.roi_widgets["FOV_Y"].widget["state"] = "disabled"

    def populate_experiment_values(self):
        """Sets values in View according to the experiment yaml file.

        Experiment yaml filed passed by controller.
        """
        self.in_initialization = True

        # Retrieve settings.
        #: dict: Camera setting dictionary
        self.camera_setting_dict = self.parent_controller.configuration["experiment"][
            "CameraParameters"
        ]
        #: dict: Microscope state dictionary
        self.microscope_state_dict = self.parent_controller.configuration["experiment"][
            "MicroscopeState"
        ]

        # Readout Settings
        self.mode_widgets["Sensor"].set(self.camera_setting_dict["sensor_mode"])
        if self.camera_setting_dict["sensor_mode"] == "Normal":
            self.mode_widgets["Readout"].set("")
            self.mode_widgets["Pixels"].set("")
        else:
            if (
                self.camera_setting_dict["readout_direction"]
                not in self.camera_readout_directions
            ):
                self.camera_setting_dict[
                    "readout_direction"
                ] = self.camera_readout_directions[0]
            self.mode_widgets["Readout"].set(
                self.camera_setting_dict["readout_direction"]
            )
            self.mode_widgets["Pixels"].set(
                self.camera_setting_dict["number_of_pixels"]
            )

        # ROI Settings
        if self.camera_setting_dict["x_pixels"] > self.default_width:
            self.camera_setting_dict["x_pixels"] = self.default_width
        if self.camera_setting_dict["y_pixels"] > self.default_height:
            self.camera_setting_dict["y_pixels"] = self.default_height
        self.roi_widgets["Width"].set(self.camera_setting_dict["x_pixels"])
        self.roi_widgets["Height"].set(self.camera_setting_dict["y_pixels"])

        # Binning settings
        self.roi_widgets["Binning"].set(self.camera_setting_dict["binning"])

        # Camera Framerate Info - 'exposure_time', 'readout_time',
        # 'framerate', 'frames_to_average'
        # Exposure time is currently for just the first active channel
        channels = self.microscope_state_dict["channels"]
        exposure_time = channels[list(channels.keys())[0]]["camera_exposure_time"]
        self.framerate_widgets["exposure_time"].set(exposure_time)
        self.framerate_widgets["frames_to_average"].set(
            self.camera_setting_dict["frames_to_average"]
        )

        # Physical Dimensions
        self.calculate_physical_dimensions()
        # readout time
        self.calculate_readout_time()

        # after initialization
        self.in_initialization = False

    def update_experiment_values(self, *args):
        """Updates experiment yaml file according to the values in View.

        Update the dictionary so that it can be combined with all the other
        sub-controllers, and then sent to the model.

        Args:
            *args: Variable length argument list.
        """
        # Camera Operation Mode
        self.camera_setting_dict["sensor_mode"] = self.mode_widgets["Sensor"].get()
        if self.camera_setting_dict["sensor_mode"] == "Light-Sheet":
            self.camera_setting_dict["readout_direction"] = self.mode_widgets[
                "Readout"
            ].get()
            self.camera_setting_dict["number_of_pixels"] = self.mode_widgets[
                "Pixels"
            ].get()
            # light-sheet doesn't support binning
            self.roi_widgets["Binning"].set("1x1")

        # Camera Binning
        self.camera_setting_dict["binning"] = self.roi_widgets["Binning"].get()

        # Camera FOV Size.
        x_pixel = self.roi_widgets["Width"].get()
        y_pixel = self.roi_widgets["Height"].get()

        # Round to nearest step
        x_pixels = int(x_pixel // self.step_width) * self.step_width
        y_pixels = int(y_pixel // self.step_height) * self.step_height

        if x_pixels < self.min_width:
            x_pixels = self.min_width
        if y_pixels < self.min_height:
            y_pixels = self.min_height

        self.camera_setting_dict["pixel_size"] = self.default_pixel_size
        self.camera_setting_dict["frames_to_average"] = self.framerate_widgets[
            "frames_to_average"
        ].get()

        binning = [
            int(x) if x != "" else 1
            for x in self.camera_setting_dict["binning"].split("x")
        ]
        img_width = x_pixels // binning[0]
        img_height = y_pixels // binning[1]

        self.camera_setting_dict["x_pixels"] = x_pixels
        self.camera_setting_dict["y_pixels"] = y_pixels
        self.camera_setting_dict["img_x_pixels"] = img_width
        self.camera_setting_dict["img_y_pixels"] = img_height

        self.roi_widgets["Width"].set(x_pixels)
        self.roi_widgets["Height"].set(y_pixels)

        return True

    def update_sensor_mode(self, *args):
        """Updates the camera sensor mode.

        Updates text in readout widget based on what sensor mode is selected
        If we are in the Light Sheet mode, then we want the camera
        self.model['CameraParameters']['sensor_mode']) == 12

        If we are in the normal mode, then we want the camera
        self.model['CameraParameters']['sensor_mode']) == 1

        Should initialize from the configuration file to the default version

        Parameters
        ----------
        *args : Variable length argument list.
        """
        # Camera Mode
        sensor_value = self.mode_widgets["Sensor"].widget.get()
        if sensor_value == "Normal":
            self.mode_widgets["Readout"].set(" ")
            self.mode_widgets["Readout"].widget["state"] = "disabled"
            self.mode_widgets["Pixels"].widget["state"] = "disabled"
            self.mode_widgets["Pixels"].widget.set("")
            self.mode_widgets["Sensor"].widget.selection_clear()

            self.show_verbose_info("Normal Camera Readout Mode")

        elif sensor_value == "Light-Sheet":
            # readout-direction from experiment
            if (
                self.camera_setting_dict["readout_direction"]
                not in self.camera_readout_directions
            ):
                self.camera_setting_dict[
                    "readout_direction"
                ] = self.camera_readout_directions[0]
            self.mode_widgets["Readout"].widget.set(
                self.camera_setting_dict["readout_direction"]
            )
            self.mode_widgets["Readout"].widget["state"] = "readonly"
            self.mode_widgets["Pixels"].set(
                self.camera_setting_dict["number_of_pixels"]
            )
            self.mode_widgets["Pixels"].widget.trigger_focusout_validation()
            self.mode_widgets["Pixels"].widget["state"] = "normal"

            self.show_verbose_info("Light Sheet Camera Readout Mode")

        # calculate readout time
        self.calculate_readout_time()

    def update_exposure_time(self, exposure_time):
        """When camera exposure time is changed, recalculate readout time

        Parameters
        ----------
        exposure_time : float
            exposure time in seconds
        """
        self.framerate_widgets["exposure_time"].set(exposure_time)
        self.calculate_readout_time()

    def update_roi(self, btn_name):
        """Update ROI width and height.

        Parameters
        ----------
        btn_name : roi button name
            width of roi in pixels: "All", 1600, 1024, 512
        """

        def handler(*args):
            if btn_name == "All":
                width = self.default_width
                height = self.default_height
            else:
                width = float(btn_name)
                height = width
                if width > self.default_width:
                    width = self.default_width
                if height > self.default_height:
                    height = self.default_height
            self.roi_widgets["Width"].set(width)
            self.roi_widgets["Height"].set(height)
            self.show_verbose_info("ROI width and height are changed to", width, height)

        return handler

    def update_fov(self, *args):
        """Recalculate fov and update the widgets: FOV_X and FOV_Y

        Parameters
        ----------
        *args : Variable length argument list.
        """
        if self.in_initialization:
            return
        self.calculate_physical_dimensions()

    def set_mode(self, mode):
        """Set widget configuration based upon imaging mode.

        This function will change state of widgets according to different mode
        'stop' mode will let the editable widget be 'normal'
        in 'live' and 'stack' mode, some widgets are disabled

        Parameters
        ----------
        mode : str
            One of 'live', 'z-stack', 'stop', 'single'
        """
        self.mode = mode
        state = "disabled" if mode != "stop" else "normal"
        state_readonly = "disabled" if mode != "stop" else "readonly"
        self.mode_widgets["Sensor"].widget["state"] = state_readonly
        if self.mode_widgets["Sensor"].get() == "Light-Sheet":
            self.mode_widgets["Readout"].widget["state"] = state_readonly
            self.mode_widgets["Pixels"].widget["state"] = (
                "normal" if mode == "live" else state
            )
        else:
            self.mode_widgets["Readout"].widget["state"] = "disabled"
            self.mode_widgets["Pixels"].widget["state"] = "disabled"
        self.framerate_widgets["frames_to_average"].widget["state"] = state
        self.roi_widgets["Width"].widget["state"] = state
        self.roi_widgets["Height"].widget["state"] = state
        self.roi_widgets["Binning"].widget["state"] = state_readonly
        for btn_name in self.roi_btns:
            self.roi_btns[btn_name]["state"] = state

    def calculate_physical_dimensions(self):
        """Calculate size of the FOV in microns.

        Calculates the size of the field of view according to the magnification of the
        system, the physical size of the pixel, and the number of pixels.
        update FOV_X and FOV_Y

        TODO: Should make sure that this is updated before we run the tiling wizard.
        Also can probably be done more elegantly in a configuration file and
        dictionary structure.
        """

        # pixel_size = self.default_pixel_size
        try:
            x_pixel = float(self.roi_widgets["Width"].get())
            y_pixel = float(self.roi_widgets["Height"].get())
        except ValueError as e:
            return

        microscope_state_dict = self.parent_controller.configuration["experiment"][
            "MicroscopeState"
        ]
        zoom = microscope_state_dict["zoom"]
        microscope_name = microscope_state_dict["microscope_name"]
        pixel_size = self.parent_controller.configuration["configuration"][
            "microscopes"
        ][microscope_name]["zoom"]["pixel_size"][zoom]

        physical_dimensions_x = x_pixel * pixel_size
        physical_dimensions_y = y_pixel * pixel_size

        self.roi_widgets["FOV_X"].set(physical_dimensions_x)
        self.roi_widgets["FOV_Y"].set(physical_dimensions_y)

    def calculate_readout_time(self):
        """Calculate camera readout time.


        TODO: Highly specific to Hamamatsu Orca Flash 4.0.
        Should find a way to pass this from the camera to here.
        This should be moved to the camera device/API,
        ideally by calling a command from the camera.
        """

        h = 9.74436e-6  # Readout timing constant
        sensor_mode = self.mode_widgets["Sensor"].get()
        if (self.readout_speed == 1) and (sensor_mode == "Normal"):
            h = 32.4812e-6
        # the ROI height 'subarray_vsize'
        vn = float(self.roi_widgets["Height"].get())
        exposure_time = float(self.framerate_widgets["exposure_time"].get())

        if sensor_mode == "Normal":
            #  Area sensor mode operation
            if self.trigger_source == 1:
                # Internal Trigger Source
                readout_time = exposure_time - ((vn / 2) * h)

            if self.trigger_active in [1, 2]:
                #  External Trigger Source
                #  Edge == 1, Level == 2
                readout_time = exposure_time - ((vn / 2) * h + 10 * h)

            elif self.trigger_active == 3:
                #  External Trigger Source
                #  Synchronous Readout == 3
                readout_time = exposure_time - ((vn / 2) * h + 5 * h)

        elif sensor_mode == "Light-Sheet":
            #  Progressive sensor mode operation
            readout_time = exposure_time - 1 / (exposure_time + (vn + 10) * h)

        # return readout_time
        self.framerate_widgets["readout_time"].set(readout_time)

    def update_number_of_pixels(self, *args):
        """Update the number of pixels in the ROI.

        In live mode, we should let the device know the number of pixels changed.

        Parameters
        ----------
        *args : tuple
            Unused

        """
        if self.mode != "live":
            return

        if self.pixel_event_id:
            self.view.after_cancel(self.pixel_event_id)
        
        pixels = self.mode_widgets["Pixels"].get()
        if pixels == "":
            return

        self.camera_setting_dict["number_of_pixels"] = int(pixels)
        # tell central controller to update model
        self.pixel_event_id = self.view.after(
            500,
            lambda: self.parent_controller.execute(
                "update_setting", "number_of_pixels"
            ),
        )

    def update_camera_device_related_setting(self):
        """Update caramera device related parameters.

        This function will update default width and height according to microscope name.

        """
        camera_config_dict = (
            self.parent_controller.configuration_controller.camera_config_dict
        )
        if camera_config_dict is None:
            return

        self.step_width = camera_config_dict.get("x_pixels_step", 4)
        self.step_height = camera_config_dict.get("y_pixels_step", 4)
        self.min_width = camera_config_dict.get("x_pixels_min", 4)
        self.min_height = camera_config_dict.get("y_pixels_min", 4)

        self.default_pixel_size = camera_config_dict["pixel_size_in_microns"]
        (
            self.default_width,
            self.default_height,
        ) = self.parent_controller.configuration_controller.camera_pixels
        self.trigger_source = camera_config_dict["trigger_source"]
        self.trigger_active = camera_config_dict["trigger_active"]
        self.readout_speed = camera_config_dict["readout_speed"]
        # framerate_widgets
        self.framerate_widgets["exposure_time"].widget.min = camera_config_dict[
            "exposure_time_range"
        ]["min"]
        self.framerate_widgets["exposure_time"].widget.max = camera_config_dict[
            "exposure_time_range"
        ]["max"]

        # roi max width and height
        self.roi_widgets["Width"].widget.config(to=self.default_width)
        self.roi_widgets["Height"].widget.config(to=self.default_height)
        if self.roi_widgets["Width"].get() > self.default_width:
            self.roi_widgets["Width"].set(self.default_width)
        if self.roi_widgets["Height"].get() > self.default_height:
            self.roi_widgets["Height"].set(self.default_height)

        # Center position
        self.roi_widgets["Center_X"].set(self.default_width / 2)
        self.roi_widgets["Center_Y"].set(self.default_height / 2)



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import logging
import platform

# Third Party Imports

# Local Imports
from navigate.controller.sub_controllers.gui_controller import GUIController
from navigate.view.custom_widgets.validation import ValidatedEntry, ValidatedSpinbox

# Logger Setup
p = __name__.split(".")[1]
logger = logging.getLogger(p)


class KeystrokeController(GUIController):
    """Keystroke controller"""

    def __init__(self, main_view, parent_controller):
        """Initialize the keystroke controller

        Parameters
        ----------
        main_view : MainView
            Main view
        parent_controller : MainController
            Main controller
        """
        super().__init__(main_view, parent_controller)

        # References to all sub frames
        #: tk.Frame: Camera View
        self.camera_view = main_view.camera_waveform.camera_tab  # Camera View

        # Multiposition Table
        #: MultipositionTable: Multiposition Table
        self.multi_table = main_view.settings.multiposition_tab.multipoint_list

        # Main view
        #: tk.Frame: Main view
        self.main_view = main_view.root
        #: tk.Notebook: Main tabs
        self.main_tabs = main_view.settings

        # Controllers for all sub frames
        #: CameraViewController: Camera View Controller
        self.camera_controller = parent_controller.camera_view_controller
        #: MultipositionTableController: Multiposition Table Controller
        self.multi_controller = parent_controller.multiposition_tab_controller
        #: StageController: Stage Controller
        self.stage_controller = parent_controller.stage_controller

        """Keystrokes for Camera View"""
        # Left Click binding
        self.camera_view.canvas.bind("<Button-1>", self.camera_controller.left_click)

        # Slider Binding
        self.camera_view.slider.bind("<Button-1>", self.camera_controller.slider_update)

        # MouseWheel Binding
        self.view.root.bind("<MouseWheel>", self.view.scroll_frame.mouse_wheel)
        self.camera_view.canvas.bind(
            "<Enter>", self.camera_controller_mouse_wheel_enter
        )
        self.camera_view.canvas.bind(
            "<Leave>", self.camera_controller_mouse_wheel_leave
        )

        # Right Click Binding
        if platform.system() == "Darwin":
            self.camera_view.canvas.bind(
                "<Button-2>", self.camera_controller.popup_menu
            )
        else:
            self.camera_view.canvas.bind(
                "<Button-3>", self.camera_controller.popup_menu
            )

        """Keystrokes for MultiTable"""
        #: MultiPositionTable: Multiposition Table
        self.mp_table = self.multi_table.pt
        self.mp_table.rowheader.bind(
            "<Double-Button-1>", self.multi_controller.handle_double_click
        )

        """Keystrokes for Main Window"""
        self.main_view.bind(
            "<Control-KeyRelease-j>", self.stage_controller.joystick_button_handler
        )
        self.main_view.bind("<Control-Key-1>", self.switch_tab)
        self.main_view.bind("<Control-Key-2>", self.switch_tab)
        self.main_view.bind("<Control-Key-3>", self.switch_tab)
        self.main_view.bind("<Control-Key-4>", self.switch_tab)
        self.main_view.bind_all("<Control-Key-z>", self.widget_undo)
        self.main_view.bind_all("<Control-Key-y>", self.widget_redo)

    def camera_controller_mouse_wheel_enter(self, event):
        """Mouse wheel binding for camera view

        Parameters
        ----------
        event : tkinter event
            Mouse wheel event

        Example
        -------
        >>> self.camera_view.canvas.bind("<Enter>",
        >>> self.camera_controller_mouse_wheel_enter)
        """
        self.view.root.unbind("<MouseWheel>")  # get rid of scrollbar mousewheel
        if platform.system() != "Linux":
            self.camera_view.canvas.bind(
                "<MouseWheel>", self.camera_controller.mouse_wheel
            )
        else:
            self.camera_view.canvas.bind(
                "<Button-4>", self.camera_controller.mouse_wheel
            )
            self.camera_view.canvas.bind(
                "<Button-5>", self.camera_controller.mouse_wheel
            )

    def camera_controller_mouse_wheel_leave(self, event):
        """Mouse wheel binding for camera view

        Parameters
        ----------
        event : tkinter event
            Mouse wheel event

        Example
        -------
        >>> self.camera_view.canvas.bind("<Leave>",
        >>> self.camera_controller_mouse_wheel_leave)
        """

        if platform.system() != "Linux":
            self.camera_view.canvas.unbind("<MouseWheel>")
        else:
            self.camera_view.canvas.unbind("<Button-4>")
            self.camera_view.canvas.unbind("<Button-5>")
        self.view.root.bind(
            "<MouseWheel>", self.view.scroll_frame.mouse_wheel
        )  # reinstate scrollbar mousewheel

    def switch_tab(self, event):
        """Switches between tabs

        Parameters
        ----------
        event : tkinter event
            Tab key event

        Example
        -------
        >>> self.main_view.bind("<Control-Key-1>", self.switch_tab)
        """

        key_val = int(event.keysym)
        if (key_val > 0) and (self.main_tabs.index("end") >= key_val):
            self.main_tabs.select(key_val - 1)

    def widget_undo(self, event):
        """Undo widget changes

        Parameters
        ----------
        event : tkinter event
            Undo key event

        Example
        -------
        >>> self.main_view.bind_all('<Control-Key-z>', self.widget_undo)
        """
        if isinstance(event.widget, ValidatedEntry) or isinstance(
            event.widget, ValidatedSpinbox
        ):  # Add all widgets that you want to be able to undo here
            event.widget.undo(event)

    def widget_redo(self, event):
        """Redo widget changes

        Parameters
        ----------
        event : tkinter event
            Redo key event

        Example
        -------
        >>> self.main_view.bind_all('<Control-Key-y>', self.widget_redo)
        """

        if isinstance(event.widget, ValidatedEntry) or isinstance(
            event.widget, ValidatedSpinbox
        ):  # Add all widgets that you want to be able to undo here
            event.widget.redo(event)



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only
# (subject to the limitations in the disclaimer below)
# provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Standard Library Imports
import tkinter as tk

# Third Party Imports

# Local Imports
from navigate.view.custom_widgets.popup import PopUp
from navigate.view.main_window_content.display_notebook import CameraTab

# p = __name__.split(".")[1]
# logger = logging.getLogger(p)


class CameraViewPopupWindow:
    """Popup window with waveform parameters for galvos, remote focusing, etc."""

    def __init__(self, root, microscope_name, *args, **kwargs):
        """Initialize the CameraViewPopupWindow class.

        Parameters
        ----------
        root : tkinter.Tk
            Root window of the application.
        microscope_name : str
            Name of the microscope.
        args : list
            List of arguments.
        kwargs : dict
            Dictionary of keyword arguments.
        """
        # Creating popup window with this name and size/placement, PopUp is a
        # Toplevel window
        #: PopUp: Popup window for the camera view.
        self.popup = PopUp(
            root,
            f"{microscope_name} Additional Camera View",
            "+320+180",
            top=False,
            transient=False,
        )
        self.popup.resizable(tk.TRUE, tk.TRUE)

        # Storing the content frame of the popup, this will be the parent of
        # the widgets
        content_frame = self.popup.get_frame()
        content_frame.columnconfigure(0, pad=5)
        content_frame.columnconfigure(1, pad=5)
        content_frame.rowconfigure(0, pad=5)
        content_frame.rowconfigure(1, pad=5)
        content_frame.rowconfigure(2, pad=5)

        # Formatting
        tk.Grid.columnconfigure(content_frame, "all", weight=1)
        tk.Grid.rowconfigure(content_frame, "all", weight=1)

        #: dict: Dictionary of all the input widgets.
        self.inputs = {}
        #: dict: Dictionary of all the buttons.
        self.buttons = {}

        #: CameraTab: Camera view tab.
        self.camera_view = CameraTab(content_frame)
        self.camera_view.is_popup = True
        self.camera_view.is_docked = False
        self.camera_view.grid(row=0, column=0, sticky=tk.NSEW)

    # Getters
    def get_variables(self):
        """Get the variables tied to the widgets.

        This function returns a dictionary of all the variables that are tied to each
        widget name.

        The key is the widget name, value is the variable associated.

        Returns
        -------
        dict
            Dictionary of all the variables that are tied to each widget name.
        """
        variables = {}
        for key, widget in self.inputs.items():
            variables[key] = widget.get_variable()
        return variables

    def get_widgets(self):
        """Get the dictionary that holds the input widgets.

        This function returns the dictionary that holds the input widgets.
        The key is the widget name, value is the LabelInput class that has all the data.

        Returns
        -------
        dict
            Dictionary that holds the input widgets.
        """
        return self.inputs

    def get_buttons(self):
        """Get the dictionary that holds the buttons.

        This function returns the dictionary that holds the buttons.
        The key is the button name, value is the button.

        Returns
        -------
        dict
            Dictionary that holds the buttons.
        """
        return self.buttons



# Copyright (c) 2021-2022  The University of Texas Southwestern Medical Center.
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted for academic and research use only (subject to the
# limitations in the disclaimer below) provided that the following conditions are met:

#      * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.

#      * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.

#      * Neither the name of the copyright holders nor the names of its
#      contributors may be used to endorse or promote products derived from this
#      software without specific prior written permission.

# NO EXPRESS OR IMPLIED LICENSES TO ANY PARTY'S PATENT RIGHTS ARE GRANTED BY
# THIS LICENSE. THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

# Standard Library Imports
import logging
import traceback

# Third Party Imports

# Local Imports

p = __name__.split(".")[1]

logger = logging.getLogger(p)


class TreeNode:
    """
    TreeNode class for representing a node in a control sequence tree.

    This class represents a node in a control sequence tree used for controlling the
    behavior of a microscope or similar device. Each node has a name, associated
    functions, and properties that define its behavior in the control sequence.

    Notes:
    ------
    - This class is used to represent nodes in a control sequence tree where each node
      corresponds to a feature or action in the control sequence.

    - The `node_funcs` attribute contains a dictionary of functions associated with this
      node, which are executed in response to specific signals or events.

    - The `node_type` attribute specifies whether the node is a 'one-step' node or a
      'multi-step' node.

    - The `device_related` attribute indicates whether the node is related to device
      control.

    - The `need_response` attribute indicates whether a response is expected from this
      node when it is executed.
    """

    def __init__(
        self,
        feature_name,
        func_dict,
        *,
        node_type="one-step",
        device_related=False,
        need_response=False,
    ):
        """Initialize the TreeNode object.

        Parameters:
        ----------
        feature_name : str
            The name or identifier of the feature associated with this node.
        func_dict : dict
            A dictionary containing functions associated with this node. The dictionary
            should include keys for different signals or events and corresponding
            function references to be executed when those signals are received.
        node_type : str, optional
            The type of the node, which can be 'one-step' or 'multi-step'. Default is
            'one-step'.
        device_related : bool, optional
            A boolean indicating whether the node is related to device control. Default
            is False.
        need_response : bool, optional
            A boolean indicating whether a response is needed from this node. Default is
            False.
        """
        #: str: The name or identifier of the feature associated with this node.
        self.node_name = str(feature_name)
        #: dict: A dictionary containing functions associated with this node.
        self.node_funcs = func_dict
        #: str: The type of the node, which can be 'one-step' or 'multi-step'.
        self.node_type = node_type  # 'one-step', 'multi-step'
        #: bool: A boolean indicating whether the node is related to device control.
        self.device_related = device_related
        #: bool: A boolean indicating whether a response is needed from this node.
        self.need_response = need_response  # True, False
        #: bool: A boolean indicating whether the node has been initialized.
        self.is_initialized = False
        #: TreeNode or None: A reference to the child node if one exists, otherwise
        # None.
        self.child = None
        #: TreeNode or None: A reference to the sibling node if one exists, otherwise
        # None.
        self.sibling = None

    def set_property(self, **kwargs):
        """Set the properties of the TreeNode object using keyword arguments.

        This method allows setting or updating the properties of the TreeNode object
        using keyword arguments. The method iterates through the provided keyword-
        argument pairs and updates the corresponding attributes of the TreeNode
        object.

        Parameters:
        ----------
        **kwargs : dict
            Keyword arguments with attribute names as keys and values to set as
            values.
        """

        for key in kwargs:
            if hasattr(self, key):
                setattr(self, key, kwargs[key])


class SignalNode(TreeNode):
    """SignalNode class for representing a signal-processing node in a control sequence
    tree.

    This class represents a signal-processing node in a control sequence tree used for
    controlling the behavior of a microscope or similar device. It inherits from the
    TreeNode class and adds signal processing capabilities.

    Notes:
    ------
    - This class is used to represent signal-processing nodes in a control sequence
      tree where each node corresponds to a feature or action in the control sequence.

    - The `node_funcs` attribute contains a dictionary of functions associated with this
      node, which are executed in response to specific signals or events.

    - The `node_type` attribute specifies whether the node is a 'one-step' node or a
      'multi-step' node.

    - The `device_related` attribute indicates whether the node is related to device
      control.

    - The `need_response` attribute indicates whether a response is expected from this
      node when it is executed.

    - The `wait_response` attribute indicates whether the node is currently waiting for
      a response, typically set to True after running the `main` function and before
      running the `main-response` function when `need_response` is True.
    """

    def __init__(
        self,
        feature_name,
        func_dict,
        *,
        node_type="one-step",
        device_related=False,
        need_response=False,
        **kwargs,
    ):
        """Initialize the SignalNode object.

        Parameters:
        ----------
        feature_name : str
            The name or identifier of the feature associated with this node.
        func_dict : dict
            A dictionary containing functions associated with this node. The dictionary
            should include keys for different signals or events and corresponding
            unction references to be executed when those signals are received.
        node_type : str, optional
            The type of the node, which can be 'one-step' or 'multi-step'. Default is
            'one-step'.
        device_related : bool, optional
            A boolean indicating whether the node is related to device control. Default
            is False.
        need_response : bool, optional
            A boolean indicating whether a response is needed from this node. Default is
            False.
        wait_response : bool, optional
            A boolean indicating whether the node is currently waiting for a response.
            Default is False.
        """
        super().__init__(
            feature_name,
            func_dict,
            node_type=node_type,
            device_related=device_related,
            need_response=need_response,
        )
        #: bool: A boolean indicating whether the node is currently waiting for a
        # response.
        self.wait_response = False
        if self.device_related and self.node_type == "multi-step":
            self.need_response = True

    def run(self, *args, wait_response=False):
        """Execute the main function associated with this node and handle response.

        This method is used to execute the main function associated with this node and
        handle response logic based on the node's properties. It can be used with or
        without waiting for a response.

        Parameters:
        ----------
        *args : any
            Additional arguments to pass to the main function.

        wait_response : bool, optional
            A boolean indicating whether to wait for a response from this node. Default
            is False.

        Returns:
        -------
        tuple
            A tuple containing the result of executing the main function and a boolean
            indicating whether the node execution is complete (True if complete,
            otherwise False).
        """

        # initialize the node when first time entering it
        if not self.is_initialized:
            self.node_funcs["init"]()
            self.is_initialized = True

        if not wait_response:
            result = self.node_funcs["main"](*args)
            if self.need_response:
                self.wait_response = True
                return result, False

        elif self.wait_response:
            result = self.node_funcs["main-response"](*args)
            self.wait_response = False
        elif self.device_related or self.need_response:
            return None, False
        else:
            # run(wait_response=True)
            result = self.node_funcs["main"](*args)

        if (
            self.wait_response
            or self.node_type == "multi-step"
            and not self.node_funcs["end"]()
        ):
            return result, False

        self.is_initialized = False
        return result, True


class DataNode(TreeNode):
    """DataNode class for representing a data-processing node in a control sequence
    tree.

    This class represents a data-processing node in a control sequence tree used for
    controlling the behavior of a microscope or similar device. It inherits from the
    TreeNode class and adds data processing capabilities.

    Notes:
    ------
    - This class is used to represent data-processing nodes in a control sequence tree
      where each node corresponds to a data processing step in the control sequence.

    - The `node_funcs` attribute contains a dictionary of functions associated with this
      node, which are executed during different data processing steps.

    - The `node_type` attribute specifies whether the node is a 'one-step' node or a
      'multi-step' node.

    - The `device_related` attribute indicates whether the node is related to device
      control.

    - The `need_response` attribute indicates whether a response is expected from this
      node when it is executed.

    - The `is_marked` attribute can be used to mark the node for special handling or
      to indicate its status.
    """

    def __init__(
        self,
        feature_name,
        func_dict,
        *,
        node_type="one-step",
        device_related=False,
        need_response=False,
        **kwargs,
    ):
        """Initialize the DataNode object.

        Parameters:
        ----------
        feature_name : str
            The name or identifier of the feature associated with this node.
        func_dict : dict
            A dictionary containing functions associated with this node. The dictionary
            should include keys for different data processing steps and corresponding
            function references to be executed during those steps.
        node_type : str, optional
            The type of the node, which can be 'one-step' or 'multi-step'. Default is
            'one-step'.
        device_related : bool, optional
            A boolean indicating whether the node is related to device control. Default
            is False.
        need_response : bool, optional
            A boolean indicating whether a response is needed from this node. Default is
            False.
        is_marked : bool, optional
            A boolean indicating whether the node is marked. Default is False.
        """
        super().__init__(
            feature_name,
            func_dict,
            node_type=node_type,
            device_related=device_related,
            need_response=need_response,
        )
        #: bool: A boolean indicating whether the node is marked.
        self.is_marked = False

    def run(self, *args):
        """Execute the data processing functions associated with this node.

        This method is used to execute the data processing functions associated with
        this node during the control sequence. It initializes the node if not already
        initialized and determines whether the current frame meets the criteria to
        execute the main data processing function.

        Parameters:
        ----------
        *args : any
            Additional arguments to pass to the data processing functions.

        Returns:
        -------
        tuple
            A tuple containing the result of executing the main data processing
            function and a boolean indicating whether the node execution is complete
            (True if complete, otherwise False).
        """

        if self.is_marked:
            return None, True

        # initialize the node when first time entering it
        if not self.is_initialized:
            self.node_funcs["init"]()
            self.is_initialized = True

        # to decide whether it is the target frame
        if not self.node_funcs["pre-main"](*args):
            return False, False

        result = self.node_funcs["main"](*args)

        if self.node_type == "multi-step" and not self.node_funcs["end"]():
            return result, False

        # erase flag when exit the node
        self.is_initialized = False
        return result, True


class Container:
    """Container class for managing a control sequence tree.

    This class is responsible for managing a control sequence tree, which consists of
    nodes representing different actions or steps in the control sequence. It provides
    methods for resetting the tree and performing cleanup operations on specified nodes.

    Notes:
    ------
    - The `Container` class is used to manage the execution of a control sequence
      tree. It maintains the state of the control sequence and provides methods
      for resetting the state and performing cleanup operations.

    - The `root` attribute represents the root node of the control sequence tree,
      which serves as the entry point for executing the control sequence.

    - The `curr_node` attribute tracks the currently running node in the control
      sequence.

    - The `end_flag` attribute is used to signal the termination of control sequence
      execution.

    - The `cleanup_list` attribute is a list of nodes that contain 'cleanup'
      functions. These functions are executed when the container is closed, allowing
      for resource cleanup.

    - The `is_closed` attribute indicates whether the container has been closed or not.
    """

    def __init__(self, root=None, cleanup_list=[]):
        """Initialize the Container object.

        Parameters:
        ----------
        root : TreeNode or None, optional
            The root node of the control sequence tree. Default is None.

        cleanup_list : list of TreeNode, optional
            A list of nodes containing 'cleanup' functions to be executed when the
            container is closed. Default is an empty list.
        """
        #: TreeNode or None: The root node of the control sequence tree.
        self.root = root

        #: TreeNode or None: The currently running node in the control sequence.
        self.curr_node = None

        #: bool: A flag indicating whether the control sequence execution should stop.
        self.end_flag = False

        #: list of TreeNode: A list of nodes containing 'cleanup' functions to be
        # executed when the container is closed.
        self.cleanup_list = cleanup_list

        #: bool: A boolean indicating whether the container is closed.
        self.is_closed = False

    def reset(self):
        """Reset the container's state, including the current node and end flag.

        This method resets the state of the container, including setting the current
        node to None and clearing the end flag. It prepares the container for running
        the control sequence from the beginning.
        """

        self.curr_node = None
        self.end_flag = False

    def cleanup(self):
        """Execute 'cleanup' functions for specified nodes in the tree.

        This method executes the 'cleanup' functions associated with nodes in the
        `cleanup_list`. These functions are typically used to perform resource
        cleanup or finalization when the container is closed. The 'cleanup_list'
        specifies which nodes should have their 'cleanup' functions executed.
        """

        for node in self.cleanup_list:
            try:
                node.node_funcs["cleanup"]()
            except Exception:
                pass
        self.is_closed = True


class SignalContainer(Container):
    """SignalContainer class for managing signal-based control sequences.

    This class is a specialized container for managing control sequences that consist
    of signal-based nodes. It extends the functionality of the base `Container` class
    and provides methods for running signal-based control sequences.

    Notes:
    ------
    - The `SignalContainer` class is designed for managing control sequences that are
      signal-based, typically used in applications involving devices and sensors.

    - It extends the functionality of the base `Container` class and provides methods
      for running signal-based control sequences.

    - The `number_of_execution` parameter specifies how many times the control
      sequence should be executed. The `remaining_number_of_execution` attribute keeps
      track of the remaining executions.
    """

    def __init__(self, root=None, cleanup_list=[], number_of_execution=1):
        """Initialize the SignalContainer object.

        Parameters:
        ----------
        root : TreeNode or None, optional
            The root node of the control sequence tree. Default is None.
        cleanup_list : list of TreeNode, optional
            A list of nodes containing 'cleanup' functions to be executed when the
            container is closed. Default is an empty list.
        number_of_execution : int, optional
            The number of times the control sequence should be executed. Default is 1.
        """
        super().__init__(root, cleanup_list)

        #: int: The total number of times the control sequence should be executed.
        self.number_of_execution = number_of_execution

        #: int: The remaining number of executions of the control sequence.
        self.remaining_number_of_execution = number_of_execution

    def reset(self):
        """Reset the container's state, including the current node and end flag.

        This method resets the state of the container, including setting the current
        node to None, clearing the end flag, and resetting the remaining number of
        executions. It prepares the container for running the control sequence from
        the beginning.
        """

        super().reset()
        self.remaining_number_of_execution = self.number_of_execution

    def run(self, *args, wait_response=False):
        """Run the signal-based control sequence.

        This method runs the signal-based control sequence by iterating through its
        nodes and executing them in order. It manages the execution of nodes and
        handles transitions between nodes based on their logic.

        Parameters:
        -----------
        *args : arguments
            Optional arguments to pass to the control sequence nodes.

        wait_response : bool, optional
            A flag indicating whether to wait for responses during execution.
            Default is False.

        Notes:
        ------
        - The `run` method is responsible for executing the control sequence nodes in
          the order defined by the control sequence tree.

        - It handles transitions between nodes, waits for responses if necessary, and
          tracks the remaining executions of the control sequence.
        """

        if self.end_flag or not self.root:
            self.end_flag = True
            return
        if not self.curr_node:
            self.curr_node = self.root
        while self.curr_node:
            logger.debug(f"running signal node: {self.curr_node.node_name}")
            try:
                result, is_end = self.curr_node.run(*args, wait_response=wait_response)
            except Exception:
                logger.debug(f"SignalContainer - {traceback.format_exc()}")
                self.end_flag = True
                self.cleanup()
                return
            if not is_end:
                return
            if result and self.curr_node.child:
                logger.debug(f"Signal running child of {self.curr_node.node_name} ")
                self.curr_node = self.curr_node.child
            elif self.curr_node.sibling:
                self.curr_node = self.curr_node.sibling
            else:
                self.curr_node = None
                if self.remaining_number_of_execution > 0:
                    self.remaining_number_of_execution -= 1
                    self.end_flag = self.end_flag or (
                        self.remaining_number_of_execution == 0
                    )
                return

            if self.curr_node.device_related:
                return


class DataContainer(Container):
    """DataContainer class for managing data-based control sequences.

    This class is a specialized container for managing control sequences that involve
    data-based nodes. It extends the functionality of the base `Container` class and
    provides methods for running data-based control sequences.

    Notes:
    ------
    - The `DataContainer` class is designed for managing control sequences that are
      data-based, typically used in applications involving data processing and
      analysis.

    - It extends the functionality of the base `Container` class and provides methods
      for running data-based control sequences.

    - The `run` method is responsible for executing the data-based control sequence
      nodes in the order defined by the control sequence tree. It handles transitions
      between nodes and manages node cleanup when necessary.
    """

    def __init__(self, root=None, cleanup_list=[]):
        """Initialize the DataContainer object.

        Parameters:
        ----------
        root : TreeNode or None, optional
            The root node of the control sequence tree. Default is None.
        cleanup_list : list of TreeNode, optional
            A list of nodes containing 'cleanup' functions to be executed when the
            container is closed. Default is an empty list.
        """
        super().__init__(root, cleanup_list)

    def run(self, *args):
        """Run the data-based control sequence.

        This method runs the data-based control sequence by iterating through its
        nodes and executing them in order. It manages the execution of nodes, handles
        transitions between nodes, and performs node cleanup when necessary.

        Parameters:
        -----------
        *args : arguments
            Optional arguments to pass to the control sequence nodes.

        Notes:
        ------
        - The `run` method is responsible for executing the data-based control
          sequence nodes in the order defined by the control sequence tree.

        - It handles transitions between nodes, waits for responses if necessary,
          and performs cleanup for nodes marked as not needing a response and having
          a 'one-step' node type.
        """

        if self.end_flag or not self.root:
            return
        if not self.curr_node:
            self.curr_node = self.root
        while self.curr_node:
            try:
                result, is_end = self.curr_node.run(*args)
            except Exception:
                logger.debug(f"DataContainer - {traceback.format_exc()}")
                if (
                    self.curr_node.need_response is False
                    and self.curr_node.node_type == "one-step"
                ):
                    try:
                        logger.debug(
                            f"Datacontainer cleanup node {self.curr_node.node_name}"
                        )
                        self.curr_node.node_funcs.get("cleanup", dummy_func)()
                    except Exception:
                        logger.debug(
                            f"The node({self.curr_node.node_name}) is not closed "
                            f"correctly! Please check the cleanup function"
                        )
                        pass
                    self.curr_node.is_marked = True
                    result, is_end = False, True
                else:
                    # terminate the container.
                    # the signal container may stuck there waiting a response,
                    # the cleanup function of that node should give it a fake
                    # response to make it stop
                    self.end_flag = True
                    self.cleanup()
                    return
            if not is_end:
                return
            if result and self.curr_node.child:
                self.curr_node = self.curr_node.child
            elif self.curr_node.sibling:
                self.curr_node = self.curr_node.sibling
            else:
                self.curr_node = None
                self.end_flag = True
                return

            if self.curr_node.device_related or self.curr_node.need_response:
                return


def get_registered_funcs(feature_module, func_type="signal"):
    """Get a dictionary of registered functions for a feature module.

    This function retrieves a dictionary of registered functions for a given feature
    module and function type (signal or data). It ensures that essential functions
    such as 'init', 'main', and 'end' are present in the dictionary, and assigns them
    default dummy functions if they are missing.

    Parameters:
    ----------
    feature_module : module
        The feature module containing the configuration table of functions.

    func_type : str, optional
        The type of functions to retrieve, either 'signal' (default) or 'data'.

    Returns:
    -------
    dict
        A dictionary of registered functions for the specified function type.

    Notes:
    ------
    - This function is typically used to retrieve a dictionary of functions defined
      in a feature module's configuration table. The dictionary includes 'init',
      'main', and 'end' functions, and optionally 'pre-main' for data functions and
      'main-response' for signal functions.

    - If any of the essential functions is missing in the configuration table, this
      function assigns default dummy functions to ensure the dictionary is complete.

    - The `dummy_func` function is used as a default for functions that do nothing,
      and `dummy_True` is used as a default for functions that return True.
    """

    func_dict = feature_module.config_table.get(func_type, {})

    if "init" not in func_dict:
        func_dict["init"] = dummy_func
    if "main" not in func_dict:
        func_dict["main"] = dummy_True
    if "end" not in func_dict:
        func_dict["end"] = dummy_True
    if func_type == "data" and "pre-main" not in func_dict:
        func_dict["pre-main"] = dummy_True
    if func_type == "signal" and "main-response" not in func_dict:
        func_dict["main-response"] = dummy_True
    return func_dict


def load_features(model, feature_list):
    """Load and organize a list of feature modules into a child-sibling tree structure.

    This function takes a list of feature modules and organizes them into a
    child-sibling tree structure, creating SignalNodes and DataNodes for each
    feature module. It also handles shared variables and cleanup functions for
    the nodes.

    Parameters:
    ----------
    model : object
        The model or system to which the features are applied.

    feature_list : list
        A list of dictionaries or tuples representing the feature modules and
        their configurations.

    Returns:
    -------
    SignalContainer
        A container containing the root of the SignalNode tree and a cleanup list
        for signal nodes.

    DataContainer
        A container containing the root of the DataNode tree and a cleanup list for
        data nodes.

    Notes:
    ------
    - This function creates SignalNodes and DataNodes for each feature module in
      the feature_list, organizing them into a child-sibling tree structure.

    - Shared variables are handled by creating or updating variables in the
      `shared_variables` dictionary.

    - Cleanup functions are added to the cleanup lists for signal and data nodes
      as needed.

    - The nested functions `create_node` and `build_feature_tree` assist in
      creating and organizing the nodes.

    - This function returns SignalContainer and DataContainer instances containing
      the root nodes and cleanup lists for signal and data nodes.

    - SignalNodes represent signal processing tasks, and DataNodes represent data
      processing tasks within the feature modules.
    """

    signal_cleanup_list, data_cleanup_list = [], []
    shared_variables = {}

    def create_node(feature_dict):
        """Create SignalNode and DataNode instances for a feature module.

        This function takes a feature dictionary, creates SignalNode and DataNode
        instances, and handles node configurations, including 'need_response' and
        'device_related' properties.

        Parameters:
        ----------
        feature_dict : dict
            A dictionary containing the feature module's name, arguments, and
            node configuration.

        Returns:
        -------
        SignalNode
            A SignalNode instance for the feature module.

        DataNode
            A DataNode instance for the feature module.

        Notes:
        ------
        - This function creates SignalNode and DataNode instances based on the
          feature dictionary's information.

        - It handles configurations for 'need_response' and 'device_related' based
          on the feature module's signal and data functions.

        - The created nodes are returned for inclusion in the tree structure.
        """

        args = ()
        if "args" in feature_dict:
            args = list(feature_dict["args"])
            for i, arg in enumerate(args):
                if type(arg) is dict and arg.get("type", None) == "shared_list":
                    variable_name = arg.get("name", "temp")
                    if variable_name not in shared_variables:
                        shared_variables[variable_name] = list(arg["value"])
                    args[i] = shared_variables[variable_name]
        feature = feature_dict["name"](model, *args)

        node_config = feature.config_table.get("node", {})
        # if signal function has a waiting func,
        # then the nodes are 'need_response' nodes
        if "main-response" in feature.config_table.get("signal", {}):
            node_config["need_response"] = True
            node_config["device_related"] = True
        if "node" in feature_dict:
            for k, v in feature_dict["node"].items():
                node_config[k] = v
        # 'multi-step' must set to be 'device_related'
        if node_config.get("node_type", "") == "multi-step":
            node_config["device_related"] = True

        signal_node = SignalNode(
            feature_dict["name"].__name__,
            get_registered_funcs(feature, "signal"),
            **node_config,
        )
        data_node = DataNode(
            feature_dict["name"].__name__,
            get_registered_funcs(feature, "data"),
            **node_config,
        )

        if "cleanup" in feature.config_table.get("signal", {}):
            signal_cleanup_list.append(signal_node)
        if "cleanup" in feature.config_table.get("data", {}):
            data_cleanup_list.append(data_node)

        return signal_node, data_node

    def build_feature_tree(feature_list, continue_list, break_list):
        """Build a child-sibling tree structure from a feature list.

        This function recursively builds a child-sibling tree structure from a list
        of feature modules. It creates SignalNode and DataNode instances for each
        feature module and organizes them into the tree structure.

        Parameters:
        ----------
        feature_list : list
            A list of dictionaries and tuples representing feature modules and their
            configurations.

        Returns:
        -------
        SignalNode
            The root SignalNode of the tree structure.

        DataNode
            The root DataNode of the tree structure.

        SignalNode
            The last SignalNode in the tree structure.

        DataNode
            The last DataNode in the tree structure.

        Notes:
        ------
        - This function recursively builds a child-sibling tree structure for the
          given feature list.

        - It returns the root SignalNode, root DataNode, last SignalNode, and last
          DataNode in the tree structure.

        - The last SignalNode and last DataNode are used to track the previous
          nodes for creating sibling relationships.
        """

        signal_root, data_root = None, None
        pre_signal, pre_data = None, None
        pre_signal2, pre_data2 = None, None
        for temp in feature_list:
            break_to_next = True
            signal_tail2 = None
            data_tail2 = None
            if type(temp) is dict:
                signal_head, data_head = create_node(temp)
                signal_tail = signal_head
                data_tail = data_head
                break_loop_list1, break_loop_list2 = [], []
                if "true" in temp:
                    if temp["true"] in ["break", '"break"', "'break'"]:
                        break_list.append(["child", signal_head, data_head, 1])
                    elif temp["true"] in ["continue", '"continue"', "'continue'"]:
                        continue_list.append(("child", signal_head, data_head))
                    elif temp["true"]:
                        (
                            signal_child_head,
                            data_child_head,
                            signal_child_tail,
                            data_child_tail,
                        ) = build_feature_tree(
                            temp["true"], continue_list, break_loop_list1
                        )
                        signal_head.child = signal_child_head
                        data_head.child = data_child_head
                        signal_tail = signal_child_tail
                        data_tail = data_child_tail
                        break_to_next = False
                if "false" in temp:
                    if temp["false"] in ["break", '"break"', "'break'"]:
                        break_list.append(["sibling", signal_head, data_head, 1])
                    elif temp["false"] in ["continue", '"continue"', "'continue'"]:
                        continue_list.append(("sibling", signal_head, data_head))
                    elif temp["false"]:
                        (
                            signal_sibling_head,
                            data_sibling_head,
                            signal_sibling_tail,
                            data_sibling_tail,
                        ) = build_feature_tree(
                            temp["false"], continue_list, break_loop_list2
                        )
                        signal_head.sibling = signal_sibling_head
                        data_head.sibling = data_sibling_head
                        if signal_tail == signal_head:
                            signal_tail = signal_sibling_tail
                            data_tail = data_sibling_tail
                        else:
                            signal_tail2 = signal_sibling_tail
                            data_tail2 = data_sibling_tail
                        break_to_next = False
                break_list += break_loop_list1
                break_list += break_loop_list2
            elif type(temp) is tuple:
                for node in break_list:
                    node[-1] += 1
                continue_loop_list = []
                signal_head, data_head, signal_tail, data_tail = build_feature_tree(
                    temp, continue_loop_list, break_list
                )
                signal_head.device_related = True
                data_head.device_related = True
                signal_tail.child = signal_head
                data_tail.child = data_head
                for node in continue_loop_list:
                    if node[0] == "child":
                        node[1].child = signal_tail
                        node[2].child = data_tail
                    else:
                        node[1].sibling = signal_tail
                        node[2].sibling = data_tail
                for node in break_list:
                    node[-1] -= 1
                break_to_next = False
            else:
                signal_head, data_head, signal_tail, data_tail = build_feature_tree(
                    temp, continue_list, break_list
                )
            if break_to_next:
                for i in range(len(break_list) - 1, -1, -1):
                    node = break_list[i]
                    if node[-1] <= 0:
                        if node[0] == "child":
                            node[1].child = signal_head
                            node[2].child = data_head
                        else:
                            node[1].sibling = signal_head
                            node[2].sibling = data_head
                        break_list.pop()
            # handle "true": break and "false": break
            if len(break_list) > 1 and break_list[-1][1] == pre_signal and break_list[-1][0] == pre_signal:
                continue

            if pre_signal:
                pre_signal.sibling = signal_head
                pre_data.sibling = data_head
            else:
                signal_root = signal_head
                data_root = data_head

            if pre_signal2:
                pre_signal2.sibling = signal_head
                pre_data2.sibling = data_head

            pre_signal = signal_tail
            pre_data = data_tail
            pre_signal2 = signal_tail2
            pre_data2 = data_tail2

        return signal_root, data_root, pre_signal, pre_data
    
    break_list = []
    signal_root, data_root, pre_signal, pre_data = build_feature_tree(
        feature_list, [], break_list
    )
    # add a node placeholder
    for node in break_list:
        if node[0] == "child":
            node[1].child, node[2].child = create_node({"name": DummyFeature})
    return SignalContainer(signal_root, signal_cleanup_list), DataContainer(
        data_root, data_cleanup_list
    )


def dummy_True(*args):
    """Dummy function that always returns True.

    This function serves as a placeholder for operations that need to return a
    Boolean value. It is designed to always return True.

    Parameters:
    ----------
    *args : tuple
        Variable-length argument list (unused).

    Returns:
    -------
    bool
        Always returns True.

    Notes:
    ------
    - This function is a simple placeholder that returns True and is often used in
      situations where a real function is not yet implemented or is not relevant
      for a specific context.
    """
    return True


def dummy_func(*args):
    """Dummy function with no operation.

    This function serves as a placeholder for operations that require a function
    but do not perform any actual computation. It does not have any side effects
    and simply does nothing when called.

    Parameters:
    ----------
    *args : tuple
        Variable-length argument list (unused).

    Returns:
    -------
    None

    Notes:
    ------
    - This function is used as a placeholder for situations where a function is
      expected but no real operation is needed.
    """

    pass

class DummyFeature:
    def __init__(self, model, *args):
        self.model = model
        self.config_table = {}

